'''module to estimate coagulation kernel using p. 508 of Jacobson (2005), called on by ode_gen'''

# number- but not mass-conserving coagulation code, called from ode_gen and operator-split
# from ode solver

import numpy as np 
import scipy.constants as si
from fl_reg_determ import reg_determ
from reyn_num import Reyn_num
from wk_int import W_k_int
import scipy.integrate as integ
import matplotlib.pyplot as plt
import scipy.constants as si
from mov_cen_water_eq import mov_cen_main as movcen # moving centre method for rebinning
import fullmov

def coag(RH, T, sbr, sbVi, M, rint, num_molec, num_part, tint, sbbound, rbou,
			num_comp, vdWon, rho, V0, rad0, PInit, testf, num_molec_rint, num_part_rint, 
			sbVj, coag_on, siz_str, wall_on):

	# inputs:---------------------------------------------------------
	
	# RH - relative humidity (fraction)
	# T - temperature (K)
	# sbr - size bin radius (m)
	# sbVi - single particle volume for i sizes (relating to sbr) (m3)
	# M - molecular weight of components (g/mol)	
	# rint - size(s) of interest (m)
	# num_molec - molecular concentration (molecules/cc (air)),
	# arranged by component in rows and size bins in columns, for particles in sbr
	# num_part - concentration of particles per size bin 
	# (particle/cc (air)) (columns) (excluding walls), for particles in sbr
	# tint - time interval coagulation occurs over (s)
	# sbbound - size bin volume boundaries (m3)
	# rbou - size bin radius boundaries (um)
	# num_comp - number of components
	# vdWon - flagging whether the van der Waals kernel should be calculated or ignored (0
	# for ignore, 1 for calculate)
	# rho - component densities (g/cm3) in a 1D array
	# V0 - original volume at size bin centre (um3)
	# rad0 - original radius at size bin centre (um)
	# PInit - pressure inside chamber (Pa)
	# testf - unit testing flag (0 for off, 1 for kernel testing, 2 for coagulation 
	#	testing with fixed kernel)
	# num_molec_rint - molecular concentration (molecules/cc (air)),
	# arranged by component in rows and size bins in columns, for particles in rint
	# num_part_rint - concentration of particles per size bin 
	# (particle/cc (air)) (columns) (excluding walls), for particles in rint
	# sbVj -  - single particle volume for j sizes (relating to rint) (m3)
	# coag_on - whether to allow coagulation to occur (1) or not (0)
	# siz_str - the size structure to use
	# wall_on - marker for whether wall being considered
	# --------------------------------------------------------------
	
	num_part = num_part.reshape(1, -1)
	# volume concentration of particles (m3/cc (air))
	vol_part = (sbVi*num_part).reshape(1, -1)

	# ensure sbn is integer
	sbrn = np.int(np.max(sbr.shape))
	sbn = np.int(np.max(rint.shape)) # number of size bins
	
	# call on function to determine the Knudsen no. and therefore flow 
	# regime of each size bin
	[Kni, eta_ai, rho_ai, kin_visc] = reg_determ(RH, T, sbr, PInit)
	[Knj, eta_aj, rho_aj, kin_visc] = reg_determ(RH, T, rint, PInit)

	# Reynold number and terminal fall velocity for each size bin
	[Rei, Vfi] = Reyn_num(sbr, eta_ai, rho_ai, kin_visc, 1.0e6, Kni)
	[Rej, Vfj] = Reyn_num(rint, eta_aj, rho_aj, kin_visc, 1.0e6, Knj)
	
	# repeat Knudsen number over number of size bins
	Kni_m = np.tile(Kni, (sbn, 1))
	
	# Cunningham slip-flow correction (15.30) with constant taken
	# from text below textbook equation (dimensionless)
	Gi = 1.0+Kni*(1.249+0.42*(np.exp(-0.87/Kni)))
	Gj = 1.0+Knj*(1.249+0.42*(np.exp(-0.87/Knj)))
	# particle diffusion coefficient (15.29) (m2/s)
	# multiply eta_a by 1.0e-3 to convert from g/m.s to kg/m.s
	# this makes it consistent with the units of Boltzmann constant
	Dpi = (((si.k*T)/(6.0*np.pi*sbr*(eta_ai*1.0e-3)))*Gi).reshape(sbrn, 1)
	Dpj = (((si.k*T)/(6.0*np.pi*rint*(eta_aj*1.0e-3)))*Gj).reshape(1, sbn)
	
	# repeat Dpi over size bins of j (m2/s)
	Dp_mi = np.repeat(Dpi, sbn, 1)
	# repeat Dpj over size bins of i (m2/s)
	Dp_mj = np.repeat(Dpj, sbrn, 0)
	# matrix for sums of Dp in size bins (m2/s)
	Dp_sum = Dp_mi+Dp_mj

	# zero the upper triangle part of the matrix as only the lower is needed
# 	Dp_sum[np.triu_indices(sbrn, 1, m=sbn)] = 0.0
	
	# Brownian collision kernel (m3/particle.s) (p. 508) 
	K_B = np.zeros((sbrn, sbn))

	# ensure sbr is a column array and rint is a row array
	sbr2 = np.zeros((np.max(sbr.shape), 1))
	rint2 = np.zeros((1, np.max(rint.shape)))
	sbr2[:,0] = sbr[:]
	rint2[0,:] = rint[:]
	
	# repeat size bin radii over size bins (m)
	rint2 = np.repeat(rint2, sbrn, 0)
	sbr_m = np.repeat(sbr2, sbn, 1)
	# zero the upper triangle part of the matrix as only the lower is needed
# 	rint2[np.triu_indices(sbrn, 1, m=sbn)] = 0.0
# 	sbr_m[np.triu_indices(sbrn, 1, m=sbn)] = 0.0
	
	# matrix for size bins sums (m)
	sbr_sum = sbr_m+rint2
	
	
	# size bins in continuum regime
	i = ((Kni<1.0).reshape(sbrn,1))
	# spread across rint (j)
	i = np.repeat(i, sbn, 1)
	
	# Brownian collision kernel (15.28) (m3/particle.s) in the continuum regime
	K_B[i] = 4.0*np.pi*(sbr_sum[i])*(Dp_sum[i])
	
	# size bins in free-molecular regime	
	i = ((Kni>10.0).reshape(sbrn,1))
	# spread across rint (j)
	i = np.repeat(i, sbn, 1)
	
	# single particle mass (g):
	# first, number of moles per component in a single particle (relating to sbr)
	num_mol_single_sbr = np.zeros((num_molec.shape[0], (num_part.shape[1])))
	ish = num_part[0, :]>0.
	num_mol_single_sbr[:, ish] = (num_molec[:, ish]/num_part[0, ish])/si.N_A
	# second product of number of moles and molecular weight
	weight_compon = num_mol_single_sbr*M
	# final sum mass across components for single particle mass (g)
	Mpi = np.sum(weight_compon, 0)
	# number of moles per component in a single particle (relating to rint)
	num_mol_single_rint = np.zeros((num_molec_rint.shape[0], (num_part_rint.shape[1])))
	ish = num_part_rint[0, :]>0.	
	num_mol_single_rint[:, ish] = (num_molec_rint[:, ish]/num_part_rint[0, ish])/si.N_A
	# second product of number of moles and molecular weight
	weight_compon = num_mol_single_rint*M
	Mpj = np.sum(weight_compon, 0)

	# thermal speed of particle (15.32) (m/s) (multiply mass by 1.0e-3 to 
	# convert from g to kg and therefore be consistent with Boltzmann's 
	# constant (1.380658e-23kgm2/s2.K.molec))
	Mpi = np.pi*(Mpi*1.0e-3).reshape(sbrn, 1).repeat(sbn, 1)
	nu_pi = np.zeros((sbrn, sbn))
	nu_pi[Mpi>0.] = ((8.0*si.k*T)/Mpi[Mpi>0.])**0.5
	
	Mpj = np.pi*(Mpj*1.0e-3).reshape(1, sbn).repeat(sbrn, 0)
	nu_pj = np.zeros((sbrn, sbn))
	nu_pj[Mpj>0.] = ((8.0*si.k*T)/Mpj[Mpj>0.])**0.5	

	# sum of squares of speeds (m2) (15.31)
	nu_p_sum = nu_pi**2.0+nu_pj**2.0
			
	# Brownian collision kernel (15.31) (m3/particle.s)
	K_B[i] = np.pi*(sbr_sum[i]**2.0)*(nu_p_sum[i]**0.5)
	
	# size bins in transition regime, where the transition regime corresponds to particles
	# of size i, not size j
	i = (1.0<=Kni) 
	j = (Kni<=10.0) # size bins in transition regime		
	i = (i*j).reshape(sbrn, 1)	# size bins in transition regime
	# spread across rint (j)
	i = np.repeat(i, sbn, 1)
	
	# particle mean free path (15.34) (m)
	lam_pi = np.zeros((len(nu_pi[:,0])))
	ish = nu_pi[:, 0]>0.
	lam_pi[ish] = ((8.0*Dpi[ish, 0])/(np.pi*nu_pi[ish, 0]))
	lam_pj = np.zeros((len(nu_pj[0,:])))
	ish = nu_pj[0, :]>0.
	lam_pj[ish] = ((8.0*Dp_mj[0, ish])/(np.pi*nu_pj[0, ish]))

	# mean distance from centre of a sphere travelled by particles
	# leaving sphere's surface and travelling lam_p (m) (15.34)
	num = (2.0*sbr+lam_pi)**3.0-(4.0*sbr**2.0+lam_pi**2.0)**1.5
	den = (6.0*sbr*lam_pi)-2.0*sbr
	sig_pi = np.zeros((len(sbr), 1))
	
	ish = den > 0.
	
	sig_pi[ish, 0] = num[ish]/den[ish]	
	sig_pi = sig_pi.repeat(sbn, 1)
	
	num = (2.0*rint+lam_pj)**3.0-(4.0*rint**2.0+lam_pj**2.0)**1.5
	den = (6.0*rint*lam_pj)-2.0*rint
	sig_pj = np.zeros((1, len(rint)))
	ish = den>0.
	sig_pj[0, ish] = num[ish]/den[ish]
	sig_pj = sig_pj.repeat(sbrn, 0)

	# sum mean distances (m)
	sig_p_sum = sig_pi**2.0+sig_pj**2.0
		
	# kernel numerator
	K_Bnum = 4.0*np.pi*sbr_sum[i]*Dp_sum[i]

	# left term kernel denominator
	K_Blden = (sbr_sum[i]/(sbr_sum[i]+(sig_p_sum[i])**0.5))
	
	# right term kernel denominator
	num = 4.0*Dp_sum[i]
	den = ((nu_p_sum[i])**0.5)*sbr_sum[i]
	K_Brden = np.zeros((len(num)))
	ish = den>0.
	K_Brden[ish] = num[ish]/den[ish] 
	
	# collision kernel (15.33) (m3/particle.s)
	K_B[i] = (K_Bnum/(K_Blden+K_Brden))
	

	
	if (testf == 1): # testing flag on
		fig, (ax0,ax1) = plt.subplots(1, 2, figsize=(12,6))
		ax0.loglog(sbr*10**6, K_B[:,0]*10**6, label='Brownian')
		ax0.set_xlabel(r'Radius of second particle ($\rm{\mu}$m)', fontsize=10)
		ax0.set_ylabel(r'Coagulation kernel ($\rm{cm^{3}particle^{-1}s^{-1}}$)', fontsize=10)
		ax1.loglog(sbr*10**6, K_B[:,1]*10**6, label='Brownian')
		ax1.set_xlabel(r'Radius of second particle ($\rm{\mu}$m)', fontsize=10)
		ax1.set_ylabel(r'Coagulation kernel ($\rm{cm^{3}particle^{-1}s^{-1}}$)', fontsize=10)
		
	
	# Convective Brownian Diffusion Enhancement kernel:

	
	# particle Schmidt number for i and jsize bins (dimensionless) (15.36)
	Scpi = kin_visc/Dpi
	Scpj = kin_visc/Dpj

	# repeat Rej over sbrn and Rei over sbn
	Re_jm = np.repeat(Rej.reshape(1,sbn), sbrn, 0)
	Re_im = np.repeat(Rei.reshape(sbrn,1), sbn, 1)
	Scpj_m = np.repeat(Scpj.reshape(1,sbn), sbrn, 0)
	Scpi_m = np.repeat(Scpi.reshape(sbrn,1), sbn, 1)

	# repeat Reynold number and Schmidt number arrays over size bins
	i = (Re_jm<=1.0)
	j = rint2>=sbr2
	i2 = i*j # Rej less than or equal to 1, and rj greater than or equal to ri
	i3 = (Re_im<=1.0)
	j2 = rint2<sbr2
	j3 = i3*j2 # Rei less than or equal to 1 and ri greater than rj
	
	i = (Re_jm>1.0)
	i3 = i*j # Rej greater than 1 and rj greater than or equal to ri
	i4 = (Re_im>1.0)
	j4 = i4*j2 # Rei greater than 1 and ri greater than rj
	

	# convective Brownian diffusion enhancement kernel (15.35)
	K_DE = np.zeros((sbrn, sbn))
	
	# convective Brownian diffusion enhancement kernel (15.35)
	# condition for both K_DE equation is r_j>=r_i
	K_DE[i2] = (K_B[i2]*0.45*Re_jm[i2]**(1.0/3.0)*Scpi_m[i2]**(1.0/3.0))
	K_DE[j3] = (K_B[j3]*0.45*Re_im[j3]**(1.0/3.0)*Scpj_m[j3]**(1.0/3.0))
	K_DE[i3] = (K_B[i3]*0.45*Re_jm[i3]**(1.0/2.0)*Scpi_m[i3]**(1.0/3.0))
	K_DE[j4] = (K_B[j4]*0.45*Re_im[j4]**(1.0/2.0)*Scpj_m[j4]**(1.0/3.0))
	
	if testf==1:
		ax0.loglog(sbr*10**6, K_DE[:,0]*10**6, label='Diff. Enhancement')
		ax1.loglog(sbr*10**6, K_DE[:,1]*10**6, label='Diff. Enhancement')
	

	
	# Gravitational Collection Kernel:
	Ecoll = np.zeros((sbrn, sbn))
	j = rint2>=sbr2
	Ecoll[j] = ((sbr2.repeat(sbn, 1))[j])**2.0/((sbr_sum[j])**2.0)
	j = rint2<sbr2
	Ecoll[j] = (rint2[j])**2.0/((sbr_sum[j])**2.0)
	
	# Gravitational collection kernel (15.37)
	# difference in terminal fall velocities
	del_Vf = np.abs(np.repeat(Vfj.reshape(1, sbn), sbrn, 0)-Vfi.reshape(sbrn,1))
	K_GC = Ecoll*np.pi*((sbr_sum)**2.0)*del_Vf
	
	if testf==1:
		ax0.loglog(sbr*10**6, K_GC[:,0]*10**6, label='Settling')
		ax1.loglog(sbr*10**6, K_GC[:,1]*10**6, label='Settling')
	
	
	# Kernel for Turbulent Inertial Motion:

	# rate of dissipation of turbulent kinetic energy per gram of medium 
	# (m2/s3) (8.4 for a typical value (which is apparently taken 
	# from Pruppacher and Klett 1997 (p. 511 Jacobson (2005)))
	epsilon	= 5.0e-4
	# kernel for turbulent inertial motion (15.40)
	K_TI = (((np.pi*epsilon**(3.0/4.0))/(si.g*kin_visc**(1.0/4.0)))*(
		sbr_sum**2.0)*del_Vf)
		
	if testf==1:
		ax0.loglog(sbr*10**6, K_TI[:,0]*10**6, label='Turb. inertia')
		ax1.loglog(sbr*10**6, K_TI[:,1]*10**6, label='Turb. inertia')
	
	# kernel for Turbulent Shear (15.41)
	K_TS = ((8.0*np.pi*epsilon)/(15.0*kin_visc))**0.5*(sbr_sum**3.0)
	
	if testf==1:
		ax0.loglog(sbr*10**6, K_TS[:,0]*10**6, label='Turb. shear')
		ax1.loglog(sbr*10**6, K_TS[:,1]*10**6, label='Turb. shear')
	
	# -----------------------------------------------------------------
	# Van der Waals/viscous collision kernel:	
	
	# particle radius (m)
	radi = sbr2 # radii array (m)
	
	# van der Waals factor results array and particle pair 
	# Knudsen number array: radii of particle i
	# in 1st dim., radii of particle j in 2nd
	# ratios of second particle radius to first particle - this 
	# approach commented out but useful for comparing with Fig. 15.8 
	# Jacobson (2005)
	res_all = np.zeros((radi.shape[0], rint.shape[0]))
	res_Knp = np.zeros((radi.shape[0], rint.shape[0]))
	
	A_H = 200.0*(si.k*1.0e3*T) # Hamaker constant
	
	for j in range(0, rint.shape[0]):
		if vdWon == 0: # when omitting van der Waals calculation for expediency
			break 
		for i in range(0,radi.shape[0]):
	
			ri = radi[i]
			rj = rint[j]
			a = (-A_H/(6.0*si.k*1.0e3*T))*2.0*ri*rj
			a1 = (-A_H/6.0)*2.0*ri*rj
			b = (-A_H/(6.0*si.k*1.0e3*T))
			b1 = (-A_H/6.0)
			# square of sum of size bin radii (m2)
			c = (ri+rj)**2.0
			# square of difference in size bin radii (m2)
			d = (ri-rj)**2.0
			# product of radii (m2)
			e = ri*rj
			# sum of radii (m)
			f = ri+rj
			# difference of radii (m)
			g = ri-rj
	
			# define the integration in 15.44
			def integrand(x,a,b,c,d,e,f,g):
				Dterm = (1.0+((2.6*e)/(c))*((e/(f*(x-g)))**0.5)+e/(f*(x-g)))
				Ep0_1 = a/(x**2.0-c)
				Ep0_2 = a/(x**2.0-d)
				Ep0_3 = b*np.log((x**2.0-c)/(x**2.0-d))
				rterm = 1.0/(x**2.0)
				return Dterm*np.exp(Ep0_1+Ep0_2+Ep0_3)*rterm
			# define the integration in 15.43
			def integrand2(x,a1,b1,c,d,e,f,g,T):
				# terms of Ep0
				Ep0_1 = a1/(x**2.0-c)
				Ep0_2 = a1/(x**2.0-d)
				Ep0_3 = b1*np.log((x**2.0-c)/(x**2.0-d))
				Ep0 = Ep0_1+Ep0_2+Ep0_3
				# terms of first differential Ep0
				Ep1_1 = (-2.0*a1*x)/((x**2.0-c)**2.0)
				Ep1_2 = (-2.0*a1*x)/((x**2.0-d)**2.0)
				Ep1_3 = (2.0*b1*x)/(x**2.0-c)
				Ep1_4 = (-2.0*b1*x)/(x**2.0-d)
				Ep1 = Ep1_1+Ep1_2+Ep1_3+Ep1_4
				# terms of second differential Ep0
				Ep2_1 = (6.0*a1*x**4.0-4.0*a1*c*x**2.0-2.0*a1*c**2.0)/((x**2.0-c)**4.0)	
				Ep2_2 = (6.0*a1*x**4.0-4.0*a1*d*x**2.0-2.0*a1*d**2.0)/((x**2.0-d)**4.0)
				Ep2_3 = (-2.0*b1*x**2.0-2.0*b1*c)/((x**2.0-c)**2.0)
				Ep2_4 = (2.0*b1*x**2.0+2.0*b1*d)/((x**2.0-d)**2.0)
				Ep2 = Ep2_1+Ep2_2+Ep2_3+Ep2_4
		
				return (Ep1+x*Ep2)*np.exp((-1.0/(si.k*1.0e3*T))*((x/2.0)*Ep1+Ep0))*(x**2.0)
	
			# integration bounds - note both integral functions
			# fall to negligible values after (ri+rj)*1.0e2 and if
			# infinity used as the upper bound numerical issues 
			# arise, therefore use (ri+rj)*1.0e2 for upper bound
			ilu = (ri+rj)*1.0e2 # upper
			ill = (ri+rj) # lower
			# integration in 15.44
			res = integ.quad(integrand, ill, ilu,args=(a,b,c,d,e,f,g),
				points=([ill*2.0]), limit=1000)
			# integration in 15.43
			res2 = integ.quad(integrand2,ill,ilu,
				args=(a1,b1,c,d,e,f,g,T),points=([ill*2.0]),limit=1000)
			# 15.44 and 15.43
			W_c = 1.0/(f*res[0])
			W_k = (-1.0/(2.0*c*(si.k*1.0e3)*T))*res2[0]
	
			# -----------------------------------------
			# particle Knudsen number calculated above
			
			# Cunningham slip-correction factor (dimensionless)
			Gi = 1.0+Kni[i]*(1.249+0.42*(np.exp(-0.87/Kni[i])))
			Gj = 1.0+Knj[j]*(1.249+0.42*(np.exp(-0.87/Knj[j])))
			# particle diffusion coefficient (15.29) (m2/s) (note the 
			# Boltzmann constant has units (kg m2)/(s2 K), so *1e3 to
			# convert to g from kg)
			Dpi = (((si.k*1.0e3)*T)/(6.0*np.pi*ri*eta_ai))*Gi
			Dpj = (((si.k*1.0e3)*T)/(6.0*np.pi*rj*eta_aj))*Gj	
			Mi = ((4.0/3.0)*np.pi*ri**3.0)*1.0e6
			Mj = ((4.0/3.0)*np.pi*rj**3.0)*1.0e6
			vbari = ((8.0*si.k*1.0e3*T)/(np.pi*Mi))**0.5 #15.32
			vbarj = ((8.0*si.k*1.0e3*T)/(np.pi*Mj))**0.5 #15.32
			# mean free path (15.34)
			lami = (8.0*Dpi)/(np.pi*vbari)
			lamj = (8.0*Dpj)/(np.pi*vbarj)
	
			Knp = ((lami**2.0+lamj**2.0)**0.5)/(ri+rj)
	
			# -----------------------------------------
			# get the van der Waals/viscous collision correction 
			# factor (15.42):		
			fac = 4.0*(Dpi+Dpj)/(((vbari**2.0+vbarj**2.0)**0.5)*(ri+rj))
			V_E = (W_c*(1.0+fac))/(1.0+(W_c/W_k)*fac)
			
			# fill in results
			res_Knp[i, j] = Knp
			res_all[i, j] = V_E
		
	
	# Van der Waals/collision coagulation kernel
	if vdWon == 0:
		K_V = K_B*(1.0-1.0) # when omitting van der Waals correction for expediency
	else:
		K_V = K_B*(res_all-1.0)

	# -------------------------------------------------------------
	
	# total coagulation kernel (m3/particle.s), with sbr in rows and rint in columns
	# eq. 15.27 of Jacobson (2005) says that the sum of kernels should be multiplied 
	# by a dimensionless coalescence efficiency.  For particles under 2um this should be
	# close to unity it says in the coalescence efficiency section.
	Beta = (K_B+K_DE+K_GC+K_TI+K_TS+K_V)
	
	if (testf == 1): # plot to compare to Fig. 15.7 of Jacobson (2005)
		ax0.loglog(sbr*10**6, Beta[:,0]*10**6, label='Total')
		ax1.loglog(sbr*10**6, Beta[:,1]*10**6, label='Total')
		plt.legend()
		plt.show()
		return()
	
	# zero beta for any size bins that have low number of particles
	ish = np.squeeze(num_part<1.0e-10)
	Beta[ish, :] = 0.
	ish = np.squeeze(num_part_rint<1.e-10)
	Beta[:, ish] = 0.
	
	# Perform coagulation, using the implicit approach of Jacobson (2005), given in 
	# eq. 15.5 

	# scale up by 1e6 to convert to cm3/particle.s from m3/particle.s
	# and therefore be consistent with particle concentrations which are 
	# (# particles/cm3 (air)).  For the production term due to coagulation below, we use
	# only the k,j coordinates in beta, where j goes as high as k-1, therefore, production
	# only uses the lower left triangle in beta.  However, the loss term below uses
	# k,j coordinates where j goes from 1 to the number of size bins
	if (coag_on == 1):
		Beta = Beta*1.e6
	if (coag_on == 0):
		Beta = Beta*0.
	if (testf == 2): # if testing the coagulation equations below, fix the kernels
		Beta = np.array(([[0.2, 0.4], [0.4, 0.5]]))
	if (testf == 3): # if comparing against Smoluchowski (1916 Fig. 4), fix kernel
		Beta = np.ones((sbn, sbn))*5.969331112806462e-10

	# matrix with volume of coagulated particles resulting from pairing of k and j 
	# particles single particle volumes of k and j (m3) 
	sbVmatk = (sbVi.reshape(-1, 1)).repeat(sbn, 1)
	sbVmatj = (sbVj.reshape(1, -1)).repeat(sbrn, 0)
	
	# combined volume of single coagulated particles (m3)
	coagV = sbVmatk+sbVmatj
	
	# matrix for number concentration at t-h in size bins repeated across rows
	num_partj = np.tile(num_part.reshape(1, -1), (sbn, 1))
	# matrix for volume concentration at t-h in size bins repeated across rows
	vol_partj = np.tile(vol_part.reshape(1, -1), (sbn, 1))
	
	# molecular concentration for k and j particles (# particles /cc (air)), components in
	# rows and size bins in columns.  j will represent the original concentration,
	# whereas molec_k will be updated during size bin loop below
	molec_k = np.zeros((num_molec.shape))
	molec_j = np.zeros((num_molec.shape))
	molec_k[:,:] = num_molec[:,:]
	molec_j[:,:] = num_molec[:,:]
	
	# use eq. 15.8 of Jacobson (2005) to estimate n_{k,t}
	# size bin loop, starting (importantly) with smallest size bin
	for sbi in range(sbn):
	
		# matrix for updated number concentrations (all those with index<sbi) are
		# concentrations at t, not t-h (concentrations spread across columns)
		# (# particles/cc (air))
		num_partk = np.tile(num_part.reshape(-1, 1), (1, sbrn))
		
		# full-moving structure - where coagulated particles are distributed between existing
		# size bins so that particle volumes at size bin centres are maintained
		#if (siz_str == 1):
		#	volind = np.zeros((coagV.shape[0], coagV.shape[1]))
			# only the coagV between this size bin and the one below, note that coagV equal
			# to this size bin accounted for below (coagV between this size bin and 
			# the one above)
		#	if (sbi > 0):
		#		coagV_sub_ind = (coagV>sbVi[0, sbi-1])*(coagV<sbVi[0, sbi])
				# fraction going to this size bin rather than one smaller
		#		frac_sub_in = (coagV-sbVi[0, sbi-1])/(sbVi[0, sbi]-sbVi[0, sbi-1])
		#		volind[coagV_sub_ind] = frac_sub_in[coagV_sub_ind] 

			# only the coagV between this size bin and the one above
		#	if (sbi < sbn-1):
		#		coagV_exc_ind = (coagV >= sbVi[0, sbi])*(coagV < sbVi[0, sbi+1])
				# fraction going to this size bin rather than one larger
		#		frac_exc_in = (sbVi[0, sbi+1]-coagV)/(sbVi[0, sbi+1]-sbVi[0, sbi])				
		#		volind[coagV_exc_ind] = frac_exc_in[coagV_exc_ind]
		#	if (sbi == sbn-1): # uppermost size bin
		#		coagV_exc_ind = (coagV >= sbVi[0, sbi])
		#		volind[coagV_exc_ind] = 1.

		if (siz_str == 0 or siz_str == 1): # moving-centre structure
			# index of k,j size bin pairs that can coagulate to give a volume that fits 
			# into k
			volind = (coagV>=sbbound[0, sbi])*(coagV<sbbound[0, sbi+1])

		# disallow number concentration gain when coagulating with itself 
		volind[sbi, :] = 0.
		volind[:, sbi] = 0.

		# using only the relevant k-j pairs in beta, number of particles from k and j
		# size bins coagulating to give new particle in sbi (# particles/cc.s).
		# This accounts for both the upper and lower diagonal of Beta, so considers
		# k-j pairs as well as j-k
		numsum_ind0 = (Beta*volind)*(num_partk*num_partj)
		# integrate over time
		numsum_ind0 = tint*numsum_ind0
		# multiply by 0.5 since 2 particles make 1
		numsum_ind0 = 0.5*numsum_ind0
		# sum to get the total rate of number of particles coagulating (from k and j) 
		# (# particles/cc.s)
		numsum0 = (numsum_ind0.sum()).sum()
		# numerator, note we use num_partj as want n_{k,t-h}
		num = num_partj[0, sbi]+numsum0
		
		#if (siz_str == 1): # full-moving structure
		#	if (sbi < sbn-1):
				# index of sbi pairs that give a volume beyond next size bin up
		#		volind = coagV[sbi, :] >= sbVi[0, sbi+1]
		#		volind = volind.astype(float)
				# lose a proportion of the number of particles from sbi coagulating 
				# with smaller particles to give a particle between sbi and sbi+1
		#		indx = coagV[sbi, :] < sbVi[0, sbi+1]
		#		volind[indx] = (coagV[sbi, indx]-sbVi[0, sbi])/(sbVi[0, sbi+1]-sbVi[0, sbi])
		#	if (sbi == sbn-1): # uppermost size bin
		#		volind = np.zeros((coagV.shape[1]))
				# lose half the number of particles of sbi coagulating with itself to 
				# give a volume in sbi
		#		volind[sbi] = 0.5

		if (siz_str == 0 or siz_str == 1): # moving-centre structure
			# particle number only from sbi when newly coagulated particles give a volume
			# outside the volume bounds of sbi
			volind = (coagV[sbi, :] >= sbbound[0, sbi+1])
			# lose half the number of particles of sbi coagulating with itself to give a 
			# volume within sbi.  Note, if uncertain about whether this should be 0.5 or
			# whether a factor 2 should be applied when self-coagulation gives a 
			# particle that fits a larger size bin, then run the test_coag to ensure 
			# that the loss from here
			# is number conservative with the gain in larger size bins
			if (coagV[sbi, sbi] < sbbound[0, sbi+1]):
				volind = volind.astype(float)
				volind[sbi] = 0.5

		# denominator, representing particle number loss from this size bin
		den_parts = Beta[sbi, :]*volind*num_partj[0, :]	
		den = 1.0+tint*(den_parts.sum())
		# updated number concentration (# particles/cc (air)) eq. 15.8 Jacobson (2005)
		num_part[0, sbi] = num/den
		
		# --------------------------------------------------------------------------------
		# change to molecular concentration part (molecules/cc (air))

		#if (siz_str == 1): # full-moving structure

		#	volind = np.zeros((coagV.shape[0], coagV.shape[1]))
			# only the coagV between this size bin and the one below, note that coagV equal
			# to this size bin accounted for below (coagV between this size bin and 
			# the one above)
		#	if (sbi > 0):
		#		coagV_sub_ind = (coagV>sbVi[0, sbi-1])*(coagV<sbVi[0, sbi])
				# fraction going to this size bin rather than one smaller
		#		frac_sub_in = (coagV-sbVi[0, sbi-1])/(sbVi[0, sbi]-sbVi[0, sbi-1])
		#		volind[coagV_sub_ind] = frac_sub_in[coagV_sub_ind] 

			# the coagV between this size bin and the one above
		#	if (sbi < sbn-1):
		#		coagV_exc_ind = (coagV >= sbVi[0, sbi])*(coagV < sbVi[0, sbi+1])
				# fraction going to this size bin rather than one larger
		#		frac_exc_in = (sbVi[0, sbi+1]-coagV)/(sbVi[0, sbi+1]-sbVi[0, sbi])				
		#		volind[coagV_exc_ind] = frac_exc_in[coagV_exc_ind]
		#	if (sbi == sbn-1): # uppermost size bin
		#		coagV_exc_ind = (coagV >= sbVi[0, sbi])
		#		volind[coagV_exc_ind] = 1.
		

		if (siz_str == 0 or siz_str == 1): # moving-centre structure
			# index of k,j size bin pairs that can coagulate to give a volume that fits 
			# into k
			volind = (coagV>=sbbound[0, sbi])*(coagV<sbbound[0, sbi+1])
			volind = volind.astype(float)		

		# consistency with particle number gain calculation above - note that we cannot
		# use N_{k,t} as this not yet known
		volind[sbi, :] = 0.
		
		# to find the number of molecules transferring from other size bins to this 
		# one do not multiply by 0.5 as needed to for number concentration
		numsum_ind = (Beta*volind)*(num_partk*num_partj)
		numsumk = ((numsum_ind.sum(axis=1))).reshape(-1, 1)
		
		# particle number concentration (# particles/cc (air)) gained by sbi from each bin
		num_contr = ((numsumk)*tint)[:, 0]
		# molecular concentration gained by sbi from each bin, note that if 
		# molec_k (new concentration) used instead of molec_j (old concentration), 
		# mass conservation issues can arise when two neighbouring size bins with very
		# different original number concentration coagulate to 
		# give a particle in a larger size bin
		ish = (num_partj[0, :] > 0.) # index of bins containing particles
		molec_contr = ((num_contr[ish]/num_partj[0, ish]).reshape(1, -1)*molec_j[:, ish]).sum(axis=1)	

		# molecular concentration loss part (molecules/cc (air)) ------------------------------
		#if (siz_str == 1): # full-moving structure
		#	volind = np.zeros((coagV.shape[1]))
		#	if (sbi < sbn-1): # can't lose molecules from largest size bin due to coagulation
				# index of sbi pairs that give a volume beyond next size bin up
		#		volind = coagV[sbi, :]>=sbVi[0, sbi+1]
				# lose a fraction of the molecular concentration of sbi coagulating 
				# with itself to give a volume in sbi
		#		if (coagV[sbi, sbi] < sbVi[0, sbi+1]):
		#			volind[sbi] = 1.-(sbVi[0, sbi+1]-coagV[sbi, sbi])/(sbVi[0, sbi+1]-sbVi[0, sbi])
		
		if (siz_str == 0 or siz_str == 1): # moving-centre structure
			# particle number only from sbi when newly coagulated particles give a volume
			# outside the current bounds, note size bin bound on uppermost size bin set very
			# high, so should be excluded from molecular concentration loss
			volind = (coagV[sbi, :] >= sbbound[0, sbi+1])
			# half loss of particle number when self-coagulation produces a particle within
			# sbi - note we are not yet at accounting for molecules lost, as still considering
			# particle number
			if (coagV[sbi, sbi] < sbbound[0, sbi+1]):
				volind = volind.astype(float)
				volind[sbi] = 0.

		num_lost_corr = num_partj[0, sbi]-(num_partj[0, sbi]/(1.+tint*sum(Beta[sbi, :]*volind*num_partj[0, :])))

		# number concentration of particle number lost from sbi, 
		# Eq. 15.8 with production term removed
		#num_lost =  num_partj[0, sbi] - num_partj[0, sbi]/(1.0+tint*((Beta[sbi, :]*volind*num_partj[0, :]).sum()))
		
		# fraction of particles lost via coagulation with each size bin
		# identify size bins where coagulation occurs
		#ish = ((Beta[sbi, :]*volind[:]*num_partj[0, :]) > 0.)
		#frac_loss = np.zeros((sbn))
		
		#frac_loss[ish] = ((Beta[sbi, ish]*volind[ish]*num_partj[0, ish])/
		#	(Beta[sbi, ish]*volind[ish]*num_partj[0, ish]).sum())
		
		# ignore any particles lost due to self-coagulation that gives a particle in sbi
		#if (coagV[sbi, sbi] < sbbound[0, sbi+1]):
		#	num_lost_corr = num_lost*(1.-frac_loss[sbi])
		#else:
		#	num_lost_corr = num_lost
		# molecular concentration represented by this loss
		# note that loss equation in GMD paper is equal to the num_lost_corr equation 
		# above and this
		# equation combined
		molec_loss = 0.
		if (num_partj[0, sbi]>0.):
			molec_loss = molec_j[:, sbi]*(num_lost_corr/num_partj[0, sbi])
		
		# new molecular concentration in sbi
		molec_k[:, sbi] += (molec_contr-molec_loss)

		if (testf == 2):
			# number of particles (not molecules) being lost
			num_part_lost = num_partj[0, sbi]-num_partj[0, sbi]/den
			# those lost from each bin
			num_part_lost_sb = num_part_lost*(den_parts/den_parts.sum())	
			print(str('total particle number concentration lost from size bin ' +str(sbi)+': '+str(num_part_lost)))
			print(str('particle number concentration lost from size bin ' +str(sbi)+' by coagulation with each size bin (in ascending order of size bin): '+str(num_part_lost_sb)))
			print(str('total particle number concentration gained by size bin '+str(sbi)+': '+str(numsum0)))
			print(str('particle number concentration gained by size bin ' +str(sbi)+' from each coagulating pair (in ascending order of size bin): '+str(numsum_ind0)))
			print(str('new particle concentrations per size bin : '+str(num_part) ))
			print(str('new molecular concentration per size bin : ' + str(molec_k)))
			print('type c return at the command line to continue')
			if (sbi == 1):
				return()
	# using new molecular concentration, calculate new dimensions per size bin 
	MV = (M[:, 0]/(rho)).reshape(num_comp, 1) # molar volume (cc/mol)
	# new volume of single particle per size bin (um3)
	ish = (num_part[0, :] > 1.0e-20) # only use size bins where particles reside
	Vnew = np.zeros((sbrn))
	Vnew[ish] = np.sum(((molec_k[:, ish]/(si.N_A*num_part[0, ish]))*MV*1.e12), 0)

	# new combined volume of all particles per size bin (um3)
	Vtot = Vnew*num_part[0, :]
	# remove particles and their corresponding component concentrations if their volume
	# is negligibly small
	negl_indx = (Vtot == 0.) # index of size bins with zero particle concentration
	num_part[0, negl_indx] = 0.
	Vnew[negl_indx] = 0.
	molec_k[:, negl_indx] = 0.
	
	# new radius per size bin (um)
	rad = ((3.0*Vnew)/(4.0*np.pi))**(1.0/3.0)
	# size bins with no particle assigned central radius
	ish = num_part[0, :] <= 1.e-20
	rad[ish] = rad0[ish]
	Vnew[ish] = V0[ish]
	
	# just want particle number concentration as an array with one dimension
	if num_part.ndim>1:
		num_part = num_part[0, :]
	
	# molecular concentrations (molecules/cc (air))
	y = molec_k.flatten(order='F')

	if (siz_str == 0): # moving centre
		(num_part, Vnew, y, rad, redt, blank, tnew) = movcen(num_part.reshape(-1, 1), 
		sbbound[0, :]*1.e18, 
		np.transpose(y.reshape(sbn, num_comp)), 
		sbn, num_comp, sbVi[0, :]*1.e18, 0.0, 0, MV*1.e12)
		# revert volume bounds to um3 from m3 before returning
		sbbound = sbbound[0, :]*1.e18

	if (siz_str == 1): # full-moving
		(Vnew, rad, y, 
		num_part, sbbound, rbou) = fullmov.fullmov(sbn, num_part.reshape(-1, 1),
 		num_comp, y, MV*1.e12, V0, sbbound[0, :]*1.e18, rbou)

	sbbound = sbbound.reshape(-1) # return to 1D array

	return(num_part, y, rad, Gi, eta_ai, Vnew, sbbound, rbou)



'''module to model particle loss to walls, called on by ode_gen'''
# the process of particle deposition to walls is reproduced here using either
# the McMurry and Rader (1985) model, or the user-defined parameters

import numpy as np
import scipy.constants as si
from scipy import integrate
import matplotlib.pyplot as plt 

def wallloss(Pn, Cn, Gi, eta_ai, Dp, MW, Varr, sbn, nc, TEMP, t, 
			inflectDp, pwl_xpre, pwl_xpro, inflectk, ChamR, Rader, testf, p_char, 
			e_field, num_asb):

	# inputs:----------------------------------------------------------
	
	# Pn - particle number concentration per size bin before time step
	# (# particle/cc (air))
	# Cn - particle phase concentration per component per size bin
	# after a time step over which gas phase reaction and 
	# gas-particle partitioning have occurred (molecules/cc (air))
	# Gi - Cunningham slip-correction factor (dimensionless)
	# eta_ai - dynamic viscosity of air (g/m.s)
	# Dp - particle diameters (m)
	# MW - component molecular weight (g/mol)
	# Varr - volume of single particles per size bin (m3)
	# sbn - number of size bins
	# nc - number of species
	# TEMP - system temperature (K)
	# t - time that wall loss occurs over (s)
	# inflectDp - particle diameter at which wall loss inflection occurs (m)
	# pwl_xpre - x value preceding inflection point
	# pwl_xpro - x value proceeding inflection point
	# inflectk - deposition rate at inflection (/s)
	# ChamR - spherical equivalent radius of chamber (below eq. 2 Charan (2018)) (m)
	# Rader - flag of whether or not to use Rader and McMurry approach
	# testf - flag of whether in testing mode or not (1 for one 0 for normal mode)
	# p_char - average number of charges per particle (/particle)
	# e_field - average electric field inside chamber (g.m/A.s3)
	# num_asb - number of actual particle size bins
	# ----------------------------------------------------------------
	if Rader == 0: # manual input of wall loss rate
		
		Beta = np.zeros((Pn.shape))
		Beta[Dp<inflectDp, 0] = 10**((np.log10(inflectDp)-
								np.log10(Dp[Dp<inflectDp]))*pwl_xpre+np.log10(inflectk))
		Beta[Dp>=inflectDp, 0] = 10**((np.log10(Dp[Dp>=inflectDp])-
								np.log10(inflectDp))*pwl_xpro+np.log10(inflectk))
		
	if Rader == 1:
		# -------------------------------------------------------------------------
		# McMurry & Rader option McMurry 1985, DOI: 10.1080/02786828508959054
		# average number of charges per particle (/particle)
		n = p_char
		# elementary charge (C==A.s) (Charan et al. 2018)
		e = 1.602e-19
		# average magnitude of electric field (g.m/A.s3) (note: V/m == kg.m/A.s3) 
		E = e_field
		
		# electrostatic migration velocity/deposition velocity 
		# (eq. 11 McMurry and Rader (1985), eq. 5 Charan (2018)) (m/s)
		ve = np.abs((n*e*Gi*E)/(3.0*np.pi*eta_ai*Dp))
		
		
		# mass of components in particles (g)
		ish = np.squeeze(Pn<1.0e-10) # only use size bins where particles available
		ve[ish] = 0.0
		mass = np.zeros((sbn-1, nc))
		ish = (Pn>1.0e-20)[:, 0] # size bins where particles present
		
		mass[ish, :] = ((Cn.reshape(sbn-1, nc)[ish, :]/Pn[ish, 0].reshape(-1, 1))/
						si.N_A)*MW.reshape(1, nc)
		# density of particles (g/m3)
		rho = np.ones((sbn-1))
		rho[ish] = np.sum(mass[ish, :], 1)/Varr[ish]
		
		
		# terminal particle settling velocity (m/s) (eq. 4 Charan (2018) also 
		# eq. 20.4 Jacobson 2005)
		# gravitational constant has units m/s2
		vs = (Dp**2.0*rho*si.g*Gi)/(18.0*eta_ai)
	
		# particle diffusion coefficient (15.29 of Jacobson (2005) and 
		# 8.73 of Seinfeld and pandis (1998)) (m2/s)
		# multiply eta_a by 1.0e-3 to convert from g/m.s to kg/m.s
		# this makes it consistent with the units of Boltzmann constant
		Dpi = (((si.k*TEMP)/(3.0*np.pi*Dp*(eta_ai*1.0e-3)))*Gi)
		
		# eddy diffusion coefficient (/s) (scalar)
		if testf == 1:
			Ke = 6.4e-3 
		else:
			Ke = 1.0
			
		# x and y terms in eq. 2 of Charan (2018)
		# x (eq. 13 McMurry (1985)) (dimensionless)
		x = (np.pi*vs)/(2.0*((Ke*Dpi)**(1.0/2.0)))
		# y (eq. 14 McMurry (1985)) (dimensionless)
		y = (np.pi*ve)/(2.0*((Ke*Dpi)**(1.0/2.0)))
		
		z1 = (x+y)
		z2 = (x-y)
		
		# empty array for Debye function results
		D1 = np.zeros(sbn-1)
		D11 = np.zeros(sbn-1)
		
		for sbi in range(sbn-1): # size bin loop
			
			if Pn[sbi]<1.0e-10: # only consider size bins with particles inside
				continue 
			
			# state integration function
			Dint = lambda t: t/(np.exp(t)-1.0)
			# integrate for first order Debye function (eq. 3 Charan (2018))
			# there is an upper limit to the variable, above which an overflow occurs in 
			# the integration, the integration reaches a threshold result for high 
			# positive numbers (>~700), so just set this threshold as the result if input 
			# too great
			if z1[sbi] > 700.0:
				a = np.array(([1.644934066848228]))
			else:
				a = integrate.quad(Dint, 0.0, z1[sbi])
			if z2[sbi] > 700.0:
				a1 = np.array(([1.644934066848228]))
			else:
				a1 = integrate.quad(Dint, 0.0, z2[sbi])
	
			D1[sbi] = (1.0/z1[sbi])*a[0]
			D11[sbi] = (1.0/z2[sbi])*a1[0]
		
		# first bit of Beta (loss rate to walls (/s)) calculation (eq. 2 Charan (2018))
		Beta1 = np.zeros((sbn-1))
		Beta1[ish] = (3.0*((Ke*Dpi[ish])**(0.5)))/(np.pi*ChamR*x[ish])
		Beta2 = ((x+y)**2.0)/2.0+(x+y)*D1+(x-y)*D11 # second bit
	
		# Beta (loss rate to walls (/s)) calculation (eq. 2 Charan (2018) and eq. 15 
		# McMurry and Rader 1985) - value
		# represents fraction of particles lost to walls every second (Beta meaning is 
		# just above eq. 1 of McMurry (1985) DOI: 10.1080/02786828508959054)
		Beta = (Beta1*Beta2).reshape(-1, 1)
	
	if (Beta<0).sum()>0:
		Beta[Beta<0] = 0.0
		print('Warning Beta in walloss.py estimated below 0, which is not possible, so value forced to 0.0')
	
	if testf == 1: # if in test mode
		return(Beta)
	
	# integrate this fraction over the time step interval to give total 
	# fraction lost over interval	
	Beta = Beta*t
	
	# find where loss of particles exceeds available number of particles and change
	# Beta to the realistic maximum
	ish = Pn<(Beta*Pn)
	Beta[ish] = 1.0
		
	# new particle number concentration and particle-phase concentrations of components
	Pn -= (Beta*Pn)
	Cn -= (Beta*Cn.reshape(sbn-1, nc)).flatten(order='C')
			
	# remove particles and their components if particle number negative
	ish = np.array((np.where(Pn<1.0e-8)))
	for i in ish[0, :]:
		Pn[i] = 0.
		Cn[nc*(i):nc*(i+1)] = 0.
	
	# prepare output
	Pn.reshape(num_asb, 1)
	
	return(Pn, Cn)



'''plots results'''
# simulation results are represented graphically

import matplotlib.pyplot as plt
from matplotlib.colors import BoundaryNorm
from matplotlib.ticker import MaxNLocator
from matplotlib.colors import LinearSegmentedColormap # for customised colormap
import matplotlib.ticker as ticker # set colormap tick labels to standard notation
import os
import user_input as ui
import retr_out
import numpy as np
import scipy.constants as si

def plotter(caller):
	
	# inputs: ------------------------------------------------------------------
	# caller - marker for whether PyCHAM (0) or tests (2) are the calling module
	# --------------------------------------------------------------------------

	# retrieve useful information from pickle file
	[sav_name, sch_name, indx_plot, Comp0] = ui.share(1)

	if (sav_name == 'default_res_name'):
		print('Default results name was used, therefore results not saved and nothing to plot')
		return()

	dir_path = os.getcwd() # current working directory
	# obtain just part of the path up to PyCHAM home directory
	for i in range(len(dir_path)):
		if dir_path[i:i+7] == 'PyCHAM':
			dir_path = dir_path[0:i+7]
			break
	# isolate the scheme name from path to scheme
	for i in range(len(sch_name)-1, 0, -1):
		if sch_name[i] == '/':
			sch_name = sch_name[i+1::]
			break
	# remove any file formats
	for i in range(len(sch_name)-1, 0, -1):
		if sch_name[i] == '.':
			sch_name = sch_name[0:i]
			break
	

	dir_path = str(dir_path+'/PyCHAM/output/'+sch_name+'/'+sav_name)

	# chamber condition ---------------------------------------------------------

	# retrieve results
	(num_sb, num_comp, Cfac, yrec, Ndry, rbou_rec, x, timehr, comp_names, 
		_, _, _, y_MV, _, wall_on, space_mode) = retr_out.retr_out(dir_path)
	
	# number of actual particle size bins
	num_asb = num_sb-wall_on

	if caller == 0:
		plt.ion() # show results to screen and turn on interactive mode

	# prepare sub-plots depending on whether particles present
	if (num_asb) == 0: # no particle size bins
		if not (indx_plot): # check whether there are any gaseous components to plot
			print('Please note no initial gas-phase concentrations were received and no particle size bins were present, therefore there is nothing for the standard plot to show')
			return()
		fig, (ax0) = plt.subplots(1, 1, figsize=(14, 7))

	if (num_asb > 0): 
		if not (indx_plot):
			print('Please note, no initial gas-phase concentrations were registered, therefore the gas-phase standard plot will not be shown')
			fig, (ax1) = plt.subplots(1, 1, figsize=(14, 7))
		else:
			fig, (ax0, ax1) = plt.subplots(2, 1, figsize=(14, 7))

		par1 = ax1.twinx() # first parasite axis
		par2 = ax1.twinx() # second parasite axis
		
		# Offset the right spine of par2.  The ticks and label have already been
		# placed on the right by twinx above.
		par2.spines["right"].set_position(("axes", 1.2))
		# Having been created by twinx, par2 has its frame off, so the line of its
		# detached spine is invisible.  First, activate the frame but make the patch
		# and spines invisible.
		make_patch_spines_invisible(par2)
		# Second, show the right spine.
		par2.spines["right"].set_visible(True)	

	if (indx_plot):
		# gas-phase concentration sub-plot ---------------------------------------------	
		for i in range(len(indx_plot)):
		
			ax0.semilogy(timehr, yrec[:, indx_plot[i]], '+',linewidth=4.0, 
						label=str(str(Comp0[i])))

		ax0.set_ylabel(r'Gas-phase concentration (ppb)', fontsize = 14)
		ax0.set_xlabel(r'Time through simulation (hours)', fontsize = 14)
		ax0.yaxis.set_tick_params(labelsize = 14, direction = 'in')
		ax0.xaxis.set_tick_params(labelsize = 14, direction = 'in')
		ax0.legend(fontsize=14)

		# find maximum and minimum of plotted concentrations for sub-plot label		
		maxy = max(yrec[:, indx_plot].flatten())
		miny = min(yrec[:, indx_plot].flatten())	
			
		ax0.text(x = timehr[0]-(timehr[-1]-timehr[0])/10., y = maxy+((maxy-miny)/10.), s='a)', size=14)

		# end of gas-phase concentration sub-plot ---------------------------------------
	
	# particle properties sub-plot --------------------------------------------------
	if (num_asb > 0): # if particles present		

		if timehr.ndim == 0: # occurs if only one time step saved
			Ndry = np.array(Ndry.reshape(1, num_asb))
		if num_asb == 1: # just one particle size bin (wall included in num_sb)
			Ndry = np.array(Ndry.reshape(len(timehr), num_asb))

		if timehr.ndim == 0: # occurs if only one time step saved
			x = np.array(x.reshape(1, num_asb))
		if num_asb == 1: # just one particle size bin (wall included in num_sb)
			x = np.array(x.reshape(len(timehr), num_asb))

		if timehr.ndim==0: # occurs if only one time step saved
			rbou_rec = np.array(rbou_rec.reshape(1, num_sb))

		
		# plotting number size distribution --------------------------------------
	
		# don't use the first boundary as it's zero, so will error when log10 taken
		log10D = np.log10(rbou_rec[:, 1::]*2.)
		if (num_asb > 1) :
			# note, can't append zero to start of log10D to cover first size bin as the log10 of the
			# non-zero boundaries give negative results due to the value being below 1, so instead
			# assume same log10 distance as the next pair
			log10D = np.append((log10D[:, 0]-(log10D[:, 1]-log10D[:, 0])).reshape(-1, 1), log10D, axis=1)
			# radius distance covered by each size bin (log10(um))
			dlog10D = (log10D[:, 1::]-log10D[:, 0:-1]).reshape(log10D.shape[0], log10D.shape[1]-1)
		if (num_asb == 1): # single particle size bin
			# assume lower radius bound is ten times smaller than upper
			dlog10D = (log10D[:, 0]-np.log10((rbou_rec[:, 1]/10.)*2.)).reshape(log10D.shape[0], 1)
			
	
		# number size distribution contours (/cc (air))
		dNdlog10D = np.zeros((Ndry.shape[0], Ndry.shape[1]))
		dNdlog10D[:, :] = Ndry[:, :]/dlog10D[:, :]
		# transpose ready for contour plot
		dNdlog10D = np.transpose(dNdlog10D)
		
		# mask the nan values so they're not plotted
		z = np.ma.masked_where(np.isnan(dNdlog10D), dNdlog10D)
	
		# customised colormap (https://www.rapidtables.com/web/color/RGB_Color.html)
		colors = [(0.60, 0.0, 0.70), (0, 0, 1), (0, 1.0, 1.0), (0, 1.0, 0.0), (1.0, 1.0, 0.0), (1.0, 0.0, 0.0)]  # R -> G -> B
		n_bin = 100  # Discretizes the colormap interpolation into bins
		cmap_name = 'my_list'
		# Create the colormap
		cm = LinearSegmentedColormap.from_list(cmap_name, colors, N=n_bin)
	
		# set contour levels
		levels = (MaxNLocator(nbins = 100).tick_values(np.min(z[~np.isnan(z)]), 
				np.max(z[~np.isnan(z)])))
	
		# associate colours and contour levels
		norm1 = BoundaryNorm(levels, ncolors = cm.N, clip=True)
		
		# contour plot with times (hours) along x axis and 
		# particle diameters (nm) along y axis
		for ti in range(len(timehr)-1): # loop through times
			p1 = ax1.pcolormesh(timehr[ti:ti+2], (rbou_rec[ti, :]*2*1e3), z[:, ti].reshape(-1, 1), cmap=cm, norm=norm1)
	
		# if logarithmic spacing of size bins specified, plot vertical axis 
		# logarithmically
		if space_mode == 'log':
			ax1.set_yscale("log")
		ax1.set_ylabel('Diameter (nm)', size = 14)
		ax1.xaxis.set_tick_params(labelsize = 14, direction = 'in')
		ax1.yaxis.set_tick_params(labelsize = 14, direction = 'in')

		# label according to whether gas-phase plot also displayed		
		if (indx_plot):
			ax1.text(x=timehr[0]-(timehr[-1]-timehr[0])/11., y = np.amax(rbou_rec*2*1e3)*1.05, s='b)', size=14)
		else:
			ax1.text(x=timehr[0]-(timehr[-1]-timehr[0])/11., y = np.amax(rbou_rec*2*1e3)*1.3, s='a)', size=14)
		ax1.set_xlabel(r'Time through simulation (hours)', fontsize=14)
		
		cb = plt.colorbar(p1, format=ticker.FuncFormatter(fmt), pad=0.25)
		cb.ax.tick_params(labelsize=14)   
		# colour bar label
		cb.set_label('dN/dlog10(D) $\mathrm{(cm^{-3})}$', size=14, rotation=270, labelpad=20)

		# ----------------------------------------------------------------------------------------
		# total particle number concentration #/cm3
	
		# include total number concentration (# particles/cc (air)) on contour plot
		# first identify size bins with radius exceeding 3nm
		# empty array for holding total number of particles
		Nvs_time = np.zeros((Ndry.shape[0]))
	
		for i in range(num_asb): # size bin loop
			# get the times when bin exceeds 3nm - might be wanted to deal with particle counter detection limits
# 			ish = x[:, i]>3.0e-3
# 			Nvs_time[ish] += Ndry[ish, i] # sum number
			Nvs_time[:] += Ndry[:, i] # sum number
	
		p3, = par1.plot(timehr, Nvs_time, '+k', label = 'N')
	
		par1.set_ylabel('N (# $\mathrm{cm^{-3})}$', size=14, rotation=270, labelpad=20) # vertical axis label
		par1.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.0e')) # set tick format for vertical axis
		par1.yaxis.set_tick_params(labelsize=14)

		# SOA mass concentration ---------------------------------------------------------------
		# array for SOA sum with time
		SOAvst = np.zeros((1, len(timehr)))
		
		final_i = 0
		# check whether water and/or core is present
		if comp_names[-2] == 'H2O': # if both present
			final_i = 2
		if comp_names[-1] == 'H2O': # if just water
			final_i = 1
		# note that the seed component is only registered in init_conc_func if initial
		# particle concentration (pconc) exceeds zero, therefore particle-phase material 
		# must be present at start of experiment (row 0 in y)
		if final_i == 0 and y[0, num_speci:(num_speci*(num_asb+1))].sum()>1.0e-10: 
			final_i = 1
		
		for i in range(num_asb): # particle size bin loop
	
			# sum of organics in condensed-phase at end of simulation (ug/m3 (air))
			
			# to replicate the SMPS results, find the volume of particles then
			# assume a density of 1.0 g/cm3
			SOAvst[0, :] += np.sum((yrec[:, ((i+1)*num_comp):((i+2)*num_comp-final_i)]/si.N_A*(y_MV[0:-final_i])*1.0e12), axis = 1)

		# log10 of maximum in SOA
		if (max(SOAvst[0, :]) > 0):
			SOAmax = int(np.log10(max(SOAvst[0, :])))
		else:
			SOAmax = 0.
		# transform SOA so no standard notation required
		SOAvst[0, :] = SOAvst[0, :]
		
	
		p5, = par2.plot(timehr, SOAvst[0, :], 'xk', label = '[secondary]')
		par2.set_ylabel(str('[secondary] ($\mathrm{\mu g\, m^{-3}})$'), rotation=270, size=16, labelpad=25)
		# set label, tick font and [SOA] vertical axis to red to match scatter plot presentation
		par2.yaxis.label.set_color('black')
		par2.tick_params(axis='y', colors='black')
		par2.spines['right'].set_color('black')
		par2.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.0e')) # set tick format for vertical axis
		par2.yaxis.set_tick_params(labelsize=16)
		par2.text((timehr)[0], max(SOAvst[0, :])/2.0, 'assumed particle density = 1.0 $\mathrm{g\, cm^{-3}}$')
		plt.legend(fontsize=14, handles=[p3, p5] ,loc=4)	

	# end of particle properties sub-plot -----------------------------------

	# save and display
	plt.savefig(str(dir_path+'/'+sav_name+'_output_plot.png'))
	if caller == 2:
		plt.show()	

	return()

# function to makes spines invisible
def make_patch_spines_invisible(ax):
	ax.set_frame_on(True)
	ax.patch.set_visible(False)
	for sp in ax.spines.values():
		sp.set_visible(False)

# function for doing colorbar tick labels in standard notation
def fmt(x, pos):
	a, b = '{:.1e}'.format(x).split('e')
	b = int(b)
	return r'${} \times 10^{{{}}}$'.format(a, b)



'''solution of ODEs for water gas-particle partitioning'''
# module to solve system of ordinary differential equations (ODEs) using solve_ivp of Scipy 

import numpy as np
import scipy.sparse as SP
from scipy.integrate import solve_ivp

# define function
def ode_solv(y, integ_step, rindx, pindx, rstoi, pstoi, 
	nreac, nprod, rrc, jac_stoi, njac, jac_den_indx, 
	jac_indx, Cinfl_now, y_arr, y_rind, uni_y_rind, 
	y_pind, uni_y_pind, reac_col, prod_col, 
	rstoi_flat, pstoi_flat, rr_arr, rr_arr_p,
	rowvals, colptrs, num_comp, num_sb,
	wall_on, Psat, Cw, act_coeff, kw, jac_wall_indx,
	seedi, core_diss, kelv_fac, kimt, num_asb,
	jac_part_indx,
	rindx_aq, pindx_aq, rstoi_aq, pstoi_aq,
	nreac_aq, nprod_aq, jac_stoi_aq, njac_aq, jac_den_indx_aq, jac_indx_aq,
	y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, uni_y_pind_aq,
	reac_col_aq, prod_col_aq, rstoi_flat_aq,
	pstoi_flat_aq, rr_arr_aq, rr_arr_p_aq, eqn_num, jac_mod_len,
	jac_part_hmf_indx, rw_indx, N_perbin, jac_part_H2O_indx, H2Oi):

	# inputs: -------------------------------------
	# y - initial concentrations (moleucles/cm3)
	# integ_step - the maximum integration time step (s)
	# rindx - index of reactants per equation
	# pindx - index of products per equation
	# rstoi - stoichiometry of reactants
	# pstoi - stoichiometry of products
	# nreac - number of reactants per equation
	# nprod - number of products per equation
	# rrc - reaction rate coefficient
	# jac_stoi - stoichiometries relevant to Jacobian
	# njac - number of Jacobian elements affected per equation
	# jac_den_indx - index of component denominators for Jacobian
	# jac_indx - index of Jacobian to place elements per equation (rows)
	# Cinfl_now - influx of components with constant influx 
	#		(molecules/cc/s)
	# y_arr - index for matrix used to arrange concentrations, 
	#	enabling calculation of reaction rate coefficients 
	# y_rind - index of y relating to reactants for reaction rate 
	# 	coefficient equation
	# uni_y_rind - unique index of reactants 
	# y_pind - index of y relating to products
	# uni_y_pind - unique index of products 
	# reac_col - column indices for sparse matrix of reaction losses
	# prod_col - column indices for sparse matrix of production gains
	# rstoi_flat - 1D array of reactant stoichiometries per equation
	# pstoi_flat - 1D array of product stoichiometries per equation
	# rr_arr - index for reaction rates to allow reactant loss
	# 	calculation
	# rr_arr_p - index for reaction rates to allow reactant loss
	# 	calculation
	# rowvals - row indices of Jacobian elements
	# colptrs - indices of  rowvals corresponding to each column of the
	# 	Jacobian
	# num_comp - number of components
	# num_sb - number of size bins
	# wall_on - flag saying whether to include wall partitioning
	# Psat - pure component saturation vapour pressures (molecules/cc)
	# Cw - effective absorbing mass concentration of wall (molecules/cc) 
	# act_coeff - activity coefficient of components
	# kw - mass transfer coefficient to wall (/s)
	# jac_wall_indx - index of inputs to Jacobian by wall partitioning
	# seedi - index of seed material
	# core_diss - dissociation constant of seed material
	# kelv_fac - kelvin factor for particles
	# kimt - mass transfer coefficient for gas-particle partitioning (s)
	# num_asb - number of actual size bins (excluding wall)
	# jac_part_indx - index for sparse Jacobian for particle influence 
	# eqn_num - number of gas- and aqueous-phase reactions 
	# jac_mod_len - modification length due to high fraction of component(s)
	# 	in particle phase
	# jac_part_hmf_indx - index of Jacobian affected by water
	#	 in the particle phase
	# rw_indx - indices of rows affected by water in particle phase
	# N_perbin - number concentration of particles per size bin (#/cc)
	# jac_part_H2O_indx - sparse Jacobian indices for the effect of
	#	particle-phase water on all other components
	# H2Oi - index for water
	# ---------------------------------------------

	def dydt(t, y): # define the ODE(s)
		
		# inputs: ----------------
		# y - water concentrations (molecules/cc), note when using scipy integrator solve_ivp, 
		#	this should have shape (number of elements, 1)
		# t - time interval to integrate over (s)
		# ---------------------------------------------
		
		# ensure y is correct shape
		if (y.shape[1] > 1):
			y = y[:, 0].reshape(-1, 1)
		# empty array to hold rate of change per component
		dd = np.zeros((y.shape[0], 1))
		
		# gas-particle partitioning-----------------
		
		# update the water particle-phase concentrations in the matrix of particle-phase 
		# concentrations for all components
		ymat[:, H2Oi] = y[1::, 0]
		
		
		# total particle-phase concentration per size bin (molecules/cc (air))
		csum = ((ymat.sum(axis=1)-ymat[:, seedi].sum(axis=1))+((ymat[:, seedi]*core_diss).sum(axis=1)).reshape(-1)).reshape(-1, 1)
		
		# size bins with contents 
		isb = (csum[:, 0] > 0.)
		
		if (any(isb)): # if particle-phase components present
			# mole fraction of water at particle surface
			Csit = (y[1::, 0][isb]/csum[isb, 0])
			# gas-phase concentration of water at particle surface (molecules/cc (air))
			Csit = Csit*Psat[isb, H2Oi]*kelv_fac[isb, 0]*act_coeff[isb, H2Oi]
			# partitioning rate (molecules/cc/s)
			dd_all = (kimt[isb, H2Oi]*(y[0, 0]-Csit)).reshape(-1, 1)
			dd[0, 0] -= sum(dd_all) # gas-phase change
			
			dd[1::, 0][isb] += (dd_all.flatten()) # particle change
			
		
		# force all components in size bins with no particle to zero
		if (num_asb > 0):
			dd[1:num_asb+1, 0][N_perbin[:, 0] == 0] = 0.
		# return to array, note that consistent with the solve_ivp manual, this ensures dd is
		# a vector rather than matrix, since y0 is a vector
		dd = dd.flatten()
		return (dd)

	def jac(t, y): # define the Jacobian
		
		# inputs: ----------------
		# y - concentrations (molecules/cc), note when using scipy integrator solve_ivp, this should have shape (number of elements, 1)
		# t - time interval to integrate over (s)
		# ---------------------------------------------
		
		# ensure y is correct shape
		if (y.ndim == 2):
			if (y.shape[1] > 1):
				y = y[:, 0].reshape(-1, 1)
		if (y.ndim <= 1):
			y = y.reshape(-1, 1)
		
		# elements of sparse Jacobian matrix
		data = np.zeros(((num_asb+1)**2-num_asb*(num_asb-1)))
		
		# the row and column indices for the sparse Jacobian
		rowvals = np.arange(num_asb+1)
		rowvals_app = np.zeros((2, num_asb))
		rowvals_app[1, :] = np.arange(1, num_asb+1)
		rowvals_app = rowvals_app.flatten(order='F')
		rowvals = np.concatenate((rowvals, rowvals_app)).astype('int')
		colptrs = np.array((0, num_asb+1))
		colptrs_app = np.arange(1, num_asb+1)*2
		colptrs = np.concatenate((colptrs, colptrs[-1]+colptrs_app))
		
		# gas-particle partitioning
		if (sum(N_perbin[:, 0]) > 0.): # if any particles present 
			data[0] = -kimt[:, H2Oi].sum(axis=0) # effect of gas on gas
		
		y[1::][N_perbin[:, 0] == 0] = 0 # ensure zero water where zero particles
		
		# update the particle-phase concentrations of water (molecules/cc (air))
		ymat[:, H2Oi] = y[1::, 0]
		# total particle-phase concentration per size bin (molecules/cc (air))
		csum = ymat.sum(axis=1)-ymat[:, seedi].sum(axis=1)+(ymat[:, seedi]*core_diss).sum(axis=1)
		
		# effect of particle-on-gas and particle-on-particle
		for isb in range(int(num_asb)): # size bin loop
			if (csum[isb] > 0): # if components present in this size bin
				# effect of gas on particle
				data[1+isb] += kimt[isb, H2Oi]
				# prepare for diagonal (component effect on itself)
				diag = kimt[isb, H2Oi]*Psat[0, H2Oi]*act_coeff[0, H2Oi]*kelv_fac[isb, 0]*(-(csum[isb]-y[1+isb, :])/(csum[isb]**2.)) 
				# implement to part_eff
				data[(num_asb+1)+isb*2] -= diag
				data[(num_asb+1)+isb*2+1] += diag

		# create Jacobian
		j = SP.csc_matrix((data, rowvals, colptrs))
		
		return(j)
	
	# set ODE solver tolerances
	atol = 0.001
	rtol = 0.0001
	
	y0 = y # remember initial component concentrations (molecules/cc (air))
	
	# isolate just the water concentrations
	y_w = y[H2Oi:num_comp*(num_asb+1):num_comp]
	
	# transform particle phase concentrations into size bins in rows and components in columns
	ymat = (y[num_comp:num_comp*(num_asb+1)]).reshape(num_asb, num_comp)
	# force all components in size bins with no particle to zero
	ymat[N_perbin[:, 0] == 0, :] = 0.
	
	# call on the ODE solver, note y contains the initial condition(s) (molecules/cc (air)) 
	# and must be 1D even though y in dydt and jac has shape (number of elements, 1)
	sol = solve_ivp(dydt, [0, integ_step], y_w, atol = atol, rtol = rtol, method = 'Radau', t_eval = [integ_step], vectorized = True, jac = jac)
	
	# force all components in size bins with no particle to zero
	y_w = np.squeeze(sol.y)
	y_w = y_w.reshape(num_asb+1, 1)
	if (num_asb > 0):
		y_w[1:num_asb+1, 0][N_perbin[:, 0] == 0] = 0.
	# return to array
	y_w = y_w.flatten()
	
	# implement new water gas- and particle-phase concentrations (molecules/cc (air))
	y0[H2Oi:num_comp*(num_asb+1):num_comp] = y_w
	# updated concentrations (molecules/cc (air))
	y[:] = y0[:]
	
	# return concentration(s) and time(s) following integration
	return(y, sol.t)



'''isolate sections of chemical scheme'''
# using the chemical scheme markers, sections of the 
# chemical scheme are separated

import re
import numpy as np
import formatting

# define function
def sch_interr(total_list_eqn, chm_sch_mrk):

	# inputs: ------------------------------------------------------------------
	# total_list_eqn - all lines from the chemical scheme file
	# chm_sch_mrk - markers to denote different section of the chemical scheme
	# --------------------------------------------------------------------------
	
	# preparatory part ---------------------------------------------------------
	eqn_list = [] # empty list for gas-phase reaction equation
	aqeqn_list = [] # empty list for particle-phase reaction equation
	RO2_names = [] # empty list for peroxy radicals
	rrc = [] # empty list for reaction rate coefficients
	rrc_name = [] # empty list for reaction rate coefficient labels
	eqn_flag = 0 # don't collate reaction equations until seen
	pr_flag = 0 # don't collate peroxy radicals until seen
	RO2_count = 0 # count on number of lines considered in peroxy radical list
	# -------------------------------------------------------------------------

	# obtain lists for reaction rate coefficients, peroxy radicals and equation reactions
	# using markers for separating chemical scheme elements
	for line in total_list_eqn:
		
		line1 = line.strip() # remove bounding white space
		
		# --------------------------------------------------------------------------------
		# generic reaction rate coefficients part
		# marker at end of generic reaction rate coefficient lines
		# the first \ allows python to interpret the second \ as a dash
		# to use in regex which means an escape in case the marker is a
		# regex special character
		# the $ means occurs at end of string
		end_mark = str('\\' + chm_sch_mrk[7]+'$')
		# look out for start of generic reaction rate coefficients
		# could be generic reaction coefficient if just one = in line
		
		if (len(line1.split('=')) == 2):
			rrc_flag = 1
			# don't record if nothing preceding '=' (can occur in KPP files, e.g.
			# =IGNORE)
			if (len((line1.split('=')[0]).strip()) == 0):
				rrc_flag = 0
			# don't record if this just an IGNORE command
			if len((line1.split('=')[1]).strip()) >= 6:
				if (line1.split('=')[1]).strip()[0:6] == 'IGNORE':
					rrc_flag = 0
			# don't record if marker (if one present) for end of generic reaction rate
			# coefficient lines not present
			if (len(chm_sch_mrk[7]) > 0):
				if re.search(end_mark, line1.strip()) == None:
					rrc_flag = 0
			
			if (rrc_flag == 1): 
				
				# dont consider if start of peroxy radical list
				if (line1.split('=')[0]).strip() != chm_sch_mrk[1]:
					# don't consider if a gas-phase chemical scheme reaction
					if ((line1.split('=')[0]).strip())[0] != chm_sch_mrk[0]:
						# don't consider if an aqueous-phase chemical scheme reaction
						if ((line1.split('=')[0]).strip())[0] != chm_sch_mrk[8]:
							# remove end characters
							line2 = line1.replace(str(chm_sch_mrk[7]), '')
							# remove all white space
							line2 = line2.replace(' ', '')
							# convert fortran-type scientific notation to python type
							line2 = formatting.SN_conversion(line2)
							# ensure rate coefficient is python readable
							line2 = formatting.convert_rate_mcm(line2)
							rrc.append(line2.strip())
							# get just name of generic reaction rate coefficient
							rrc_name.append((line2.split('=')[0]).strip())		
			
		# --------------------------------------------------------------------------------
		# peroxy radical part
		# now start logging peroxy radicals
		RO2_start_mark = str('^' + chm_sch_mrk[1])
		
		# if starting marker for peroxy radical list seen, flag that recording needed
		if (re.match(RO2_start_mark, line1) != None):
			
			# to double check that recording needed for peroxy radicals (in case 
			# chm_sch_mrk[1] is not unique
			
			# first check whether the RO2 list comprises just one line, as this will
			# mean its end marker is present
			if (len(chm_sch_mrk[5].strip()) > 0):
				# .* allows search across all elements of line, \\ ensures marker is
				# recognised as string
				mark = str('.*\\' + chm_sch_mrk[5])
				if (re.match(mark, line1) != None):
					pr_flag = 1
			
			# look for presence of marker for RO2 list continuing onto next line, which
			# confirms this is the RO2 list when it covers more than one line
			# .* allows search across all elements of line, \\ ensures marker is
			# recognised as string
			mark = str('.*\\' + chm_sch_mrk[6])
			if (re.match(mark, line1) != None):
				pr_flag = 1
			
			# if line end or continuation marker not supplied then assume the RO2 start
			# marker is unique
			if (len(chm_sch_mrk[5].strip()) == 0 and len(chm_sch_mrk[6].strip()) == 0):
				pr_flag = 1
					
		if (pr_flag == 1):
			# get the elements in line separated by peroxy radical separator
			line2 = line1.split(chm_sch_mrk[2])
			
			RO2_count += 1 # count on number of lines considered in peroxy radical list
			
			for line3 in line2: # loop through elements in line
				if len(line3.split('='))>1: # in case of RO2 = ...
					line3 = (line3.split('='))[1]
				if len(line3.split(';'))>1: # in case of RO2 list finishing with ...;
					line3 = (line3.split(';'))[0]
				if len(line3.split('&'))>1: # in case of RO2 list finishing with &
					line3 = (line3.split('&'))[0]
				
				# remove any white space
				line3 = line3.strip()
				# don't include white space or ampersands
				if (line3 == '' or line3 == '&'):
					continue
				
				else: # if not these, then first strip surrounding marks
					if line3[0:len(chm_sch_mrk[3])] == chm_sch_mrk[3]:
						line3 = line3[len(chm_sch_mrk[3])::]
					if line3[-len(chm_sch_mrk[4])::] == chm_sch_mrk[4]:
						line3 = line3[0:-len(chm_sch_mrk[4])]
					
					RO2_names.append(line3)
			# check for end of RO2 list - given either by marker for end or absence of
			# marker for continuation onto next line of RO2
			# check for marker for end of RO2 list
			if (len(chm_sch_mrk[5].strip()) > 0):
				# .* allows search across all elements of line, \\ ensures marker is
				# recognised as string
				mark = str('.*\\' + chm_sch_mrk[5])
				if (re.match(mark, line1) != None):
					pr_flag = 0
			
			else: # look for absence of marker for RO2 list continuing onto next line
				# .* allows search across all elements of line, \\ ensures marker is
				# recognised as string
				mark = str('.*\\' + chm_sch_mrk[6])
				if (re.match(mark, line1) == None):
					pr_flag = 0
		
		# --------------------------------------------------------------------------------
		# gas-phase reaction equation part
		# ^ means occurs at start of line and, first \ means second \ can be interpreted 
		# and second \ ensures recognition of marker
		marker = str('^\\' +  chm_sch_mrk[0])
		# first check is whether equation start marker is present
		if (re.match(marker, line1) != None):
			# second check is whether markers for starting reaction rate coefficients
			# part, and markers for end of equation lines, are present
			eqn_markers = [str('.*\\' +  chm_sch_mrk[9]), str('.*\\' +  chm_sch_mrk[11])]
			if (re.match(eqn_markers[0], line1) != None and 
				re.match(eqn_markers[1], line1) != None):
				eqn_list.append(line1) # store reaction equations

		# aqueous-phase reaction equation part
		# ^ means occurs at start of line and, first \ means second \ can be interpreted 
		# and second \ ensures recognition of marker
		# first, check if a marker given, if not bypass
		if chm_sch_mrk[8] == '':
			continue
		else:
			
			marker = str('^\\' +  chm_sch_mrk[8])
			if (re.match(marker, line1) != None):
				# second check is whether markers for starting reaction rate coefficients
				# part, and markers for end of equation lines, are present
				eqn_markers = [str('.*\\' +  chm_sch_mrk[9]), str('.*\\' +  chm_sch_mrk[11])]
				if (re.match(eqn_markers[0], line1) != None and 
					re.match(eqn_markers[1], line1) != None):
					aqeqn_list.append(line1) # store reaction equations

	# number of equations
	eqn_num = np.array((len(eqn_list), len(aqeqn_list)))
	
	return(eqn_list, aqeqn_list, eqn_num, rrc, rrc_name, RO2_names)



'''estimating the natural light intensity for photochemistry'''
# equivalent to the method for AtChem2

import numpy as np


def zenith(time, lat, lon, DayOfYear):
    
	# inputs ------------------------------------------------------
	# time is a float in s, represents time of day
	# lat is latitude and lon is longitude in degrees
	# DayOfYear is the day number of the year (between 1-365)
	# -------------------------------------------------------------

	pi = 4.0*np.arctan(1.0) # 1pi=180deg=3.14. For future calculation
	radian = 1.80e+02/pi # 1 unit rad; use this to convert [\deg] to [\rad]
	dec = 0.41 # solar declination angle (rad), consistent with AtChem2 default

	# latitude conversion from degrees to radian
	lat = lat/radian # in [\rad]
	
	# the day angle (radians), from calcTheta inside solarFunctions.f90 file of AtChem2,
	# assuming not a leap year (correct for 2010)
	theta = 2.*pi*DayOfYear/365.
	
	# equation of time accounts for the discrepancy between the apparent and the mean
	# solar time at a given location
	c0 = 0.000075
	c1 = 0.001868
	c2 = -0.032077
	c3 = -0.014615
	c4 = -0.040849
	eqtime = c0 + c1*np.cos(theta)+c2*np.sin(theta)+c3*np.cos(2.0*theta)+c4*np.sin(2.0*theta)

	# get the fraction hour by dividing by seconds per hour, then remove any hours
	# from previous days by using the remainder function and 24 hours
	currentFracHour = np.remainder(time/3600., 24.)
	
	# local hour angle (lha): representing the time of the day - taken from 
	# solarFunctions.f90 of AtChem2
	lha = pi*((currentFracHour/12.)-(1.+lon/180.))+eqtime

	sinld = np.sin(lat)*np.sin(dec)
	cosld = np.cos(lat)*np.cos(dec)
   
	cosx = (np.cos(lha)*cosld)+sinld
	secx = 1.E+0/(cosx+1.E-30)

	# as in solarFunctions.f90 of AtChem2, set negative cosx to 0
	if (cosx < 0.):
		cosx = 0.0
		secx = 100.0
	
	return (secx, cosx)



'''module for calculating photolysis rates''' 
# photolysis rates calculated for specified absorption cross-section 
# and quantum yield files using actinic flux file'''

import numpy as np
import os

def lamp_photo(fname, J, TEMP, act_flux_path):

	# --------------------------------------------------------------
	# inputs
	# fname - name of folder where photoloysis information stored
	# J - array of photolysis rate
	# TEMP - chamber temperature (K)
	# act_flux_path - path of actinic flux file, 'no' if none given (in which case this
	# module should not be called)
	
	# --------------------------------------------------------------
	# open wavelengths (nm) we have total actinic flux for (photon/cm2/nm/s)
	# from chamber
	f = open(act_flux_path, 'r')
	wl_chm = np.empty(0) # chamber wavelengths (nm)
	act_chm = np.empty(0) # chamber actinic flux
	
	for line in f: # loop through line
		
		try: # omit headers
			float((line.strip()).split(',')[0])
			wl = float((line.strip()).split(',')[0])
			act = float((line.strip()).split(',')[1])
			wl_chm = np.append(wl_chm, (np.array(wl)).reshape(1), axis=0)
			act_chm = np.append(act_chm, (np.array(act)).reshape(1), axis=0)
			
		except:
			continue
	f.close() # close file
	
	# determine whether to use MCM estimation for wavelength-dependent absorption 
	# cross-sections (cm2/molecule) and quantum yields (fraction) or other
	cwd = os.getcwd() # address of current working directory
	
	if fname != str(cwd+'/PyCHAM/photofiles/MCMv3.2'): # using user-supplied estimates
		 
		# open file to read
		f = open(fname, 'r')
		
		# keep count on photolysis reactions, note Fortran indexing to be consistent with 
		# MCM photochemical reaction numbers
		Ji = 1
		
		wlxs = np.empty(0) # will contain wavelengths for cross-sections (nm)
		all_xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		wlqy = np.empty(0) # will contain wavelengths for quantum yields (nm)
		all_qy = np.empty(0) # will contain quantum yields (fraction)
		
		# flags for when to record absorption cross section and quantum yields
		xs_rec = 0
		qy_rec = 0
		
		for line in f: # loop through line

			# know when end reached
			if line.strip() == str('J_'+str(Ji+1) + '_axs') or line.strip() == str('J_'+str(Ji+1) + '_qy') or line.strip() =='J_end':

				# absorption cross section (cm2/molecule) interpolation to wavelengths
				# given in actinic flux file
				all_xs = np.interp(wl_chm, wlxs, all_xs)
				# quantum yield (fraction) interpolation to wavelengths
				# given in actinic flux file
				all_qy = np.interp(wl_chm, wlqy, all_qy)
				# note, length of J set in eqn_parser.py
				
				J[Ji] = sum(all_xs*all_qy*act_chm)
				
				# reset to empty
				wlxs = np.empty(0) # wavelengths for cross-sections (nm)
				all_xs = np.empty(0) # absorption cross-sections (cm2/molecule)
				wlqy = np.empty(0) # wavelengths for quantum yields (nm)
				all_qy = np.empty(0) # quantum yields (fraction)
				
				if line.strip() =='J_end':
					continue
				
				Ji += 1 # prepare for next photochemical reaction
				xs_rec = 0
				qy_rec = 0
		
			# absorption cross sections for this reaction starts
			if line.strip() == str('J_'+str(Ji) + '_axs'):
				xs_rec = 1
				qy_rec = 0
				continue
				
			# quantum yields for this reaction starts
			if line.strip() == str('J_'+str(Ji) + '_qy'):
				xs_rec = 0
				qy_rec = 1
				continue
			
			if xs_rec == 1:
				wl = float(line.split(',')[0])
				xs = float(line.split(',')[1])
				wlxs = np.append(wlxs, (np.array(wl)).reshape(1), axis=0)
				all_xs = np.append(all_xs, (np.array(xs)).reshape(1), axis=0)
				
			
			if qy_rec == 1:
				wl = float(line.split(',')[0])
				qy = float(line.split(',')[1])
				wlqy = np.append(wlqy, (np.array(wl)).reshape(1), axis=0)
				all_qy = np.append(all_qy, (np.array(xs)).reshape(1), axis=0)
			
			
				
				
		f.close() # close file
	
	if fname == str(cwd+'/PyCHAM/photofiles/MCMv3.2'): # using MCM estimates
	
		# --------------------------------------------------------------
		# J<1> and J<2> for O3 (ozone) photolysis
		# cross-section file
		f = open(str(fname+'/O3/o3_molina86_cs.txt'), 'r')
		wlO3xs = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xsO3 = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
	
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				xs = float(line.split('	')[1])*np.exp(float(line.split('	')[2])/TEMP)
				wlO3xs = np.append(wlO3xs, (np.array(wl)).reshape(1), axis=0)
				xsO3 = np.append(xsO3, (np.array(xs)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xsO3 = np.interp(wl_chm, wlO3xs, xsO3)
		
		# same for J<1> quantum yield O3=O(1D)
		f = open(str(fname+'/O3/o3_o1d_matsumi02_qy_298.txt'), 'r')
		wlO3qy = np.empty(0) # will contain wavelengths for qy (nm)
		qyO3 = np.empty(0) # will contain qy (dimensionless fraction (0-1))
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				qy = float(line.split('	')[1])
				wlO3qy = np.append(wlO3qy, (np.array(wl)).reshape(1), axis=0)
				qyO3 = np.append(qyO3, (np.array(qy)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# quantum yield (J<1>) interpolation
		qyO3 = np.interp(wl_chm, wlO3qy, qyO3)
		
		
		# same for J<2> quantum yield O3=O
		f = open(str(fname+'/O3/o3_o3p_matsumi02_qy_298.txt'), 'r')
		wlO3Pqy = np.empty(0) # will contain wavelengths for qy (nm)
		qyO3P = np.empty(0) # will contain qy (dimensionless fraction (0-1))
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				qy = float(line.split('	')[1])
				wlO3Pqy = np.append(wlO3Pqy, (np.array(wl)).reshape(1), axis=0)
				qyO3P = np.append(qyO3P, (np.array(qy)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# quantum yield (J<2>) interpolation
		qyO3P = np.interp(wl_chm, wlO3Pqy, qyO3P)
		
		
		# photolysis rate for J<1> and J<2>(/s)
		J[1] = sum(xsO3*qyO3*act_chm)
		J[2] = sum(xsO3*qyO3P*act_chm)
		
		# --------------------------------------------------------------
		# J<3> for H2O2 (hydrogen peroxide) photolysis: H2O2 = OH + OH
		# cross-section file
		f = open(str(fname+'/H2O2/h2o2_iupac2003_cs_298.txt'), 'r')
		wlH2O2xs = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xsH2O2 = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
	
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				xs = float(line.split('	')[1])
				wlH2O2xs = np.append(wlH2O2xs, (np.array(wl)).reshape(1), axis=0)
				xsH2O2 = np.append(xsH2O2, (np.array(xs)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xsH2O2 = np.interp(wl_chm, wlH2O2xs, xsH2O2)
		
		# for quantum yield of H2O2 (J<3>), MCM site links to IUPAC recommendation
		# which is 1.0 above a wavelength of 230 nm and which states uncertainty below this.
		# This is true on 04/12/2019.
		# Therefore, here assume quantum yield of one for all wavelengths.
		# photolysis rate for J<3> (/s)
		J[3] = sum(xsH2O2*1.0*act_chm)
		
		# --------------------------------------------------------------
		# J<4> for NO2 (nitrogen dioxide) photolysis: NO2 = NO + O
		# cross-section file
		f = open(str(fname+'/NO2/no2_iupac03_cs_298.txt'), 'r')
		wlNO2xs = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xsNO2 = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split(' ')[0])
				wl = float(line.split(' ')[0])
				xs = float(line.split(' ')[1])*1.0e-20
				wlNO2xs = np.append(wlNO2xs, (np.array(wl)).reshape(1), axis=0)
				xsNO2 = np.append(xsNO2, (np.array(xs)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xsNO2 = np.interp(wl_chm, wlNO2xs, xsNO2)
		
		# same for J<4> quantum yield: NO2 = NO + O
		f = open(str(fname+'/NO2/no2_iupac03_qy_298.txt'), 'r')
		wlNO2qy = np.empty(0) # will contain wavelengths for cross-sections (nm)
		qyNO2 = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				qy = float(line.split('	')[1])
				wlNO2qy = np.append(wlNO2qy, (np.array(wl)).reshape(1), axis=0)
				qyNO2 = np.append(qyNO2, (np.array(qy)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		qyNO2 = np.interp(wl_chm, wlNO2qy, qyNO2)
		
		# photolysis rate for J<4>
		J[4] = sum(xsNO2*qyNO2*act_chm)
		
		# --------------------------------------------------------------
		# J<5> for NO3 (nitrate radical) photolysis: NO3 = NO ;
		# cross-section file
		f = open(str(fname+'/NO3/no3_iupac03_cs_298.txt'), 'r')
		wlNO3xs = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xsNO3 = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				xs = float(line.split('	')[1])*1.0e-19
				wlNO3xs = np.append(wlNO3xs, (np.array(wl)).reshape(1), axis=0)
				xsNO3 = np.append(xsNO3, (np.array(xs)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xsNO3 = np.interp(wl_chm, wlNO3xs, xsNO3)
		
		# J<5> quantum yield: NO3 = NO
		f = open(str(fname+'/NO3/no3_no_o2_johnson96_qy_298.txt'), 'r')
		wlNO3qy = np.empty(0) # will contain wavelengths for cross-sections (nm)
		qyNO3 = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				qy = float(line.split('	')[1])
				wlNO3qy = np.append(wlNO3qy, (np.array(wl)).reshape(1), axis=0)
				qyNO3 = np.append(qyNO3, (np.array(qy)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		qyNO3 = np.interp(wl_chm, wlNO3qy, qyNO3)
		
		# photolysis rate for J<5>
		J[5] = sum(xsNO3*qyNO3*act_chm)
		
		# J<6> quantum yield: NO3 = NO2 + O
		f = open(str(fname+'/NO3/no3_no2_o_johnson96_qy_298.txt'), 'r')
		wlNO3qy = np.empty(0) # will contain wavelengths for cross-sections (nm)
		qyNO3 = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				qy = float(line.split('	')[1])
				wlNO3qy = np.append(wlNO3qy, (np.array(wl)).reshape(1), axis=0)
				qyNO3 = np.append(qyNO3, (np.array(qy)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		qyNO3 = np.interp(wl_chm, wlNO3qy, qyNO3)
		
		# photolysis rate for J<6>
		J[6] = sum(xsNO3*qyNO3*act_chm)
		
		
		# --------------------------------------------------------------
		# J<7> for HONO (nitrous acid) photolysis: HONO = OH + NO ;
		# cross-section file
		f = open(str(fname+'/HONO/hono_bongartz91_94_cs_298.txt'), 'r')
		wlHONOxs = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xsHONO = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				xs = float(line.split('	')[1])
				wlHONOxs = np.append(wlHONOxs, (np.array(wl)).reshape(1), axis=0)
				xsHONO = np.append(xsHONO, (np.array(xs)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xsHONO = np.interp(wl_chm, wlHONOxs, xsHONO)
		
		# J<7> quantum yield: HONO = OH + NO
		# assume quantum yield of 1.0 for J<7> following recommendation of MCM website on
		# 4/12/2019
		
		# photolysis rate for J<7>
		J[7] = sum(xsHONO*1.0*act_chm)
		
		# --------------------------------------------------------------
		# J<8> for HNO3 (nitric acid) photolysis: HNO3 = OH + NO2;
		# cross-section file
		f = open(str(fname+'/HNO3/hno3_burkholder93_cs.txt'), 'r')
		wl_xs = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				xsA = float(line.split('	')[1])*1.0e-20
				xsB = float(line.split('	')[2])
				xss = xsA*np.exp(xsB*(TEMP-298.0))
				wl_xs = np.append(wl_xs, (np.array(wl)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl_xs, xs)
		
		# J<8> quantum yield: HNO3 = OH + NO2
		# following MCM website recommendation (true on 4/12/2019), assume 1.0
		
		# photolysis rate
		J[8] = sum(xs*1.0*act_chm)
		
		# --------------------------------------------------------------
		# J<11> for HCHO (formaldehyde) photolysis: HCHO = CO + HO2 + HO2;
		# cross-section file
		f = open(str(fname+'/HCHO/hcho_meller00_cs.txt'), 'r')
		wl_xs = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				xsA = float(line.split('	')[1])*1.0e-21
				xsB = float(line.split('	')[2])*1.0e-24
				xss = xsA+xsB*(TEMP-298.0)
				wl_xs = np.append(wl_xs, (np.array(wl)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl_xs, xs)
		
		# J<11> and J<12> quantum yield: HCHO = CO + HO2 + HO2 and HCHO = H2 + CO
		f = open(str(fname+'/HCHO/hcho_iupac03_qy_298.txt'), 'r')
		wl_qy = np.empty(0) # will contain wavelengths for cross-sections (nm)
		qy = np.empty(0) # will contain quantum yields (cm2/molecule)
		qy2 = np.empty(0) # will contain quantum yields (cm2/molecule)
		
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				qys = float(line.split('	')[1])
				qys2 = float(line.split('	')[2])
				wl_qy = np.append(wl_qy, (np.array(wl)).reshape(1), axis=0)
				qy = np.append(qy, (np.array(qys)).reshape(1), axis=0)
				qy2 = np.append(qy2, (np.array(qys2)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# quantum yield interpolation for J<11> and J<12>
		qy = np.interp(wl_chm, wl_qy, qy)
		qy2 = np.interp(wl_chm, wl_qy, qy2)
		
		# photolysis rate
		J[11] = sum(xs*qy*act_chm)
		J[12] = sum(xs*qy2*act_chm)
		
		# -------------------------------------------------------------------
		# J<13> for → CH3 + HCO
		# cross-section file
		f = open(str(fname+'/CH3CHO/ch3cho_iupac03_cs_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		
		for line in f: # loop through line
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)	
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield file
		f = open(str(fname+'/CH3CHO/ch3cho_iupac03_qy_298.txt'), 'r')
		wl_qy = np.empty(0) # will contain wavelengths for cross-sections (nm)
		qy = np.empty(0) # will contain quantum yields (cm2/molecule)
		
		for line in f: # loop through line
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				qys = float(line.split('	')[3])
				wl_qy = np.append(wl_qy, (np.array(wl)).reshape(1), axis=0)
				qy = np.append(qy, (np.array(qys)).reshape(1), axis=0)
			except:
				continue
		f.close() # close file
		# quantum yield interpolation
		qy = np.interp(wl_chm, wl_qy, qy)
		# photolysis rate
		J[13] = sum(xs*qy*act_chm)
		
		# -------------------------------------------------------------------
		# J<14> for → C2H5 + HCO
		# cross-section file
		f = open(str(fname+'/C2H5CHO/c2h5cho_iupac03_cs_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		
		for line in f: # loop through line
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)	
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield file
		f = open(str(fname+'/C2H5CHO/c2h5cho_chen&zhu01_qy_298.txt'), 'r')
		wl_qy = np.empty(0) # will contain wavelengths for cross-sections (nm)
		qy = np.empty(0) # will contain quantum yields (cm2/molecule)
		
		for line in f: # loop through line
			try: # omit headers
				float(line.split('	')[0])
				wl = float(line.split('	')[0])
				qys = float(line.split('	')[1])
				wl_qy = np.append(wl_qy, (np.array(wl)).reshape(1), axis=0)
				qy = np.append(qy, (np.array(qys)).reshape(1), axis=0)
			except:
				continue
		f.close() # close file
		# quantum yield interpolation
		qy = np.interp(wl_chm, wl_qy, qy)
		# photolysis rate
		J[14] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<15> for multiple photolysis
		# cross-section file
		f = open(str(fname+'/n_C3H7CHO/n_c3h7cho_iupac05_cs_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])*1.0e-21
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# using recommendation inside n_c3h7cho_iupac05_cs_qy_298 (true on 4/12/2019)
		qy = 0.21
		
		# photolysis rate
		J[15] = sum(xs*qy*act_chm)
		
		# -------------------------------------------------------------------
		# J<16> for → C2H4 + CH3CHO
		# cross-section file
		# use the same cross-section as used in J[15] above
		
		# using recommendation inside n_c3h7cho_iupac05_cs_qy_298 (true on 4/12/2019)
		qy = 0.10
		
		# photolysis rate
		J[16] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<17> for multiple photolysis
		# cross-section file
		f = open(str(fname+'/i_C3H7CHO/i-c3h7cho_martinez92_cs_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])*1.0e-21
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield file
		f = open(str(fname+'/i_C3H7CHO/i_c3h7cho_chen02_hco _qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths (nm)
		qy = np.empty(0) # will contain quantum yields (fractions)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				qys = float(line.split('	')[1])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				qy = np.append(qy, (np.array(qys)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# quantum yield interpolation
		qy = np.interp(wl_chm, wl, qy)
		
		# photolysis rate
		J[17] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<18> for → CH2=CCH3 + HCO and J<19> for → CH2=C(CH3)CO + H
		# cross-section file
		f = open(str(fname+'/MACR/macr_iupac05_cs_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
	
			try: # omit headers
				float(line.split('	')[0])
				float(line.split('	')[1])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield
		# for J<18> and J<19> we take the recommendation of 'macr_iupac05_cs_qy_298.txt' and 
		# use 1.95e-3 (true on 20/12/2019)
		qy = 1.95e-3
		
		# photolysis rate
		J[18] = sum(xs*qy*act_chm)
		J[19] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<20> for → CH3C(CHO)=CHCH2O + OH
		# cross-section file
		f = open(str(fname+'/C5HPALD1/C5HPALD_cs_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield
		# for J<20> take the recommendation of 'C5HPALD_cs_qy_298.txt' and 
		# use 1.0 (true on 20/12/2019)
		qy = 1.0
		
		# photolysis rate
		J[20] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<21> for CH3COCH3 (acetone) photolysis: CH3COCH3 = CH3CO3 + CH3O2;
		# cross-section file
		f = open(str(fname+'/CH3COCH3/ch3coch3_iupac05_cs.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xsA = float(line.split('	')[1])
				xsB = float(line.split('	')[2])
				xsC = float(line.split('	')[3])
				xsD = float(line.split('	')[4])
				xss = xsA*(1.0+xsB*TEMP+xsC*TEMP**2.0+xsD*TEMP**3.0)
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		f = open(str(fname+'/CH3COCH3/ch3coch3_iupac05_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths (nm)
		qy = np.empty(0) # will contain quantum yields
		for line in f: # loop through lines
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				qys = float(line.split('	')[3])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				qy = np.append(qy, (np.array(qys)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# quantum yield interpolation
		qy = np.interp(wl_chm, wl, qy)
		
		# photolysis rate
		J[21] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<22> for → CH3CO + C2H5
		# cross-section file
		f = open(str(fname+'/MEK/ch3coc2h5_iupac05_cs_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield, following recommendation from ch3coc2h5_iupac05_cs_qy_298
		# (true on 4/12/2019) to use 0.16
		qy = 0.16
		
		# photolysis rate
		J[22] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<23> for → CH3CH=CH2 + CO and J<24> for → CH3CO + CH2=CH
		# cross-section file
		f = open(str(fname+'/MVK/mvk_iupac05_cs_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield file
		f = open(str(fname+'/MVK/mvk_iupac05_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		qy = np.empty(0) # will contain quantum yields for J<23> (fraction)
		qy24 = np.empty(0) # will contain quantum yields for J<24> (fraction)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				qys = float(line.split('	')[1])
				qys24 = float(line.split('	')[2])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				qy = np.append(qy, (np.array(qys)).reshape(1), axis=0)
				qy24 = np.append(qy24, (np.array(qys24)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
	
		# quantum yield interpolation for J<23>
		qy = np.interp(wl_chm, wl, qy)
		# quantum yield interpolation for J<24>
		qy24 = np.interp(wl_chm, wl, qy24)
		
		# photolysis rate
		J[23] = sum(xs*qy*act_chm)
		J[24] = sum(xs*qy24*act_chm)
		
		# --------------------------------------------------------------
		# J<31>, J<32>, J<33> for glyoxal photolysis
		# cross-section file
		f = open(str(fname+'/CHOCHO/chocho_volkamer05_cs_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])*1.0e-20
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		f = open(str(fname+'/CHOCHO/chocho_tadic06_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths (nm)
		qy = np.empty(0) # will contain quantum yields
		qy2 = np.empty(0) # will contain quantum yields
		qy3 = np.empty(0) # will contain quantum yields
		for line in f: # loop through lines
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				qys = float(line.split('	')[1])
				qys2 = float(line.split('	')[3])
				qys3 = float(line.split('	')[2])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				qy = np.append(qy, (np.array(qys)).reshape(1), axis=0)
				qy2 = np.append(qy2, (np.array(qys2)).reshape(1), axis=0)
				qy3 = np.append(qy3, (np.array(qys3)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# quantum yield interpolation
		qy = np.interp(wl_chm, wl, qy)
		qy2 = np.interp(wl_chm, wl, qy2)
		qy3 = np.interp(wl_chm, wl, qy3)
		
		# photolysis rate
		J[31] = sum(xs*qy*act_chm)
		J[32] = sum(xs*qy2*act_chm)
		J[33] = sum(xs*qy3*act_chm)
		
		# --------------------------------------------------------------
		# J<34> for → CH3CO + HCO
		# cross-section file
		f = open(str(fname+'/CH3COCHO/ch3cocho_jpl_iupac05_cs_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])*1.0e-20
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield
		f = open(str(fname+'/CH3COCHO/ch3cocho_iupac05_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths (nm)
		qy = np.empty(0) # will contain quantum yields
	
		for line in f: # loop through lines
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				qys = float(line.split('	')[2])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				qy = np.append(qy, (np.array(qys)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# quantum yield interpolation
		qy = np.interp(wl_chm, wl, qy)
		
		# photolysis rate
		J[34] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<35> for → CH3CO + CH3CO
		# cross-section file
		f = open(str(fname+'/BIACET/biacet_horowitz01_cs_plum83_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])*1.0e-20
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# following quantum yield value given in biacet_horowitz01_cs_plum83_qy_298.txt
		# (true on 4/12/2019) 
		qy = 0.158
		
		# photolysis rate
		J[35] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<41> for → CH3O + OH
		# cross-section file
		f = open(str(fname+'/CH3OOH/ch3ooh_iupac05_cs_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])*1.0e-20
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# follow quantum yield value given in ch3ooh_iupac05_cs_qy_298.txt
		# (true on 4/12/2019)
		qy = 1.0
		
		# photolysis rate
		J[41] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<51> for → CH3O + NO2
		# cross-section file
		f = open(str(fname+'/CH3ONO2/ch3ono2_iupac05_cs_qy.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])*1.0e-20
				xssB = float(line.split('	')[2])*1.0e-3
				xss = xss*np.exp(xssB*(TEMP-298.0))
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# follow quantum yield value given in ch3ono2_iupac05_cs_qy.txt
		# (true on 4/12/2019)
		qy = 1.0
		
		# photolysis rate
		J[51] = sum(xs*qy*act_chm)
	
		# --------------------------------------------------------------
		# J<52> for → C2H5O + NO2
		# cross-section file
		f = open(str(fname+'/CH3CH2ONO2/ch3ch2ono2_iupac05_cs_qy.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])
				xssB = float(line.split('	')[2])
				xss = xss*np.exp(xssB*(TEMP-298.0))
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# follow quantum yield value given in ch3ch2ono2_iupac05_cs_qy.txt
		# (true on 4/12/2019)
		qy = 1.0
		
		# photolysis rate
		J[52] = sum(xs*qy*act_chm)
	
		# --------------------------------------------------------------
		# J<53> for → n-C3H7O + NO2
		# cross-section file
		f = open(str(fname+'/NOA/noa_barnes93_cs_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield
		f = open(str(fname+'/NOA/noa_estimated_qy_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths (nm)
		qy = np.empty(0) # will contain quantum yields
	
		for line in f: # loop through lines
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				qys = float(line.split('	')[1])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				qy = np.append(qy, (np.array(qys)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# quantum yield interpolation
		qy = np.interp(wl_chm, wl, qy)
		
		# photolysis rate
		J[53] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<54> for → CH3C(O.)CH3 + NO2
		# cross-section file
		f = open(str(fname+'/CH3CH2ONO2/ch3ch2ono2_iupac05_cs_qy.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss1 = float(line.split('	')[1])
				xss2 = float(line.split('	')[2])
				xss = xss1 #+xss2*(TEMP-298.0)
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield uses value given in ch3ch2ono2_iupac05_cs_qy.txt
		# (true on 4/12/2019)
		qy=1.0
		
		# photolysis rate
		J[54] = sum(xs*qy*act_chm)
	
		# --------------------------------------------------------------
		# J<55> for → t-C4H9O + NO2
		# comparing the link given in the MCM site: http://mcm.leeds.ac.uk/MCMv3.3.1/parameters/photolysis/t_C4H9ONO2/t_c4h9ono2_Roberts&Fajar89_cs_298.txt
		# for tert-butyl nitrate to the download folder of photolysis rates,
		# this process does not seem to be included in the download, so manual input provided
		wl = np.array(([270.0, 275.0, 280.0, 285.0, 290.0, 295.0, 300.0, 305.0, 310.0, 315.0, 320.0, 325.0, 330.0]))
		xs = np.array(([4.3e-20, 4.0e-20, 3.7e-20, 3.1e-20, 2.6e-20, 2.0e-20, 1.5e-20, 1.0e-20, 7.0e-21, 4.5e-21, 2.7e-21, 1.50e-21, 8.60e-22]))
		
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield uses value given in http://mcm.leeds.ac.uk/MCMv3.3.1/parameters/photolysis/t_C4H9ONO2/t_c4h9ono2_Roberts&Fajar89_cs_298.txt
		# (true on 4/12/2019)
		qy=1.0
		
		# photolysis rate
		J[55] = sum(xs*qy*act_chm)
		
		# --------------------------------------------------------------
		# J<56> for → CH3C(O)CH2(O.) + NO2 and → CH3CO + HCHO + NO2
		# cross-section file
		f = open(str(fname+'/NOA/noa_barnes93_cs_298.txt'), 'r')
		wl = np.empty(0) # will contain wavelengths for cross-sections (nm)
		xs = np.empty(0) # will contain absorption cross-sections (cm2/molecule)
		for line in f: # loop through line
			
			try: # omit headers
				float(line.split('	')[0])
				wls = float(line.split('	')[0])
				xss = float(line.split('	')[1])
				wl = np.append(wl, (np.array(wls)).reshape(1), axis=0)
				xs = np.append(xs, (np.array(xss)).reshape(1), axis=0)
				
			except:
				continue
		f.close() # close file
		# absorption cross section (cm2/molecule) interpolation
		xs = np.interp(wl_chm, wl, xs)
		
		# quantum yield uses value given in MCM website table (http://mcm.leeds.ac.uk/MCMv3.3.1/parameters/photolysis.htt)
		# (true on 4/12/2019)
		qy=0.9
		
		# photolysis rate
		J[56] = sum(xs*qy*act_chm)
	
	return J



'''module to calculate gas-phase concentration of water'''
# based on input relative humidity and temperature the
# concentration of water in the gas-phase is estimated

import numpy as np

def water_calc(TEMP, RH, NA):

	# saturation vapour pressure (Pa) (TEMP in K.)
	Psat_water = (np.exp((-0.58002206E4/TEMP)+0.13914993E1-(0.48640239E-1*TEMP) 
		+ (0.41764768E-4*(TEMP**2.0E0))-(0.14452093E-7*(TEMP**3.0E0))+ 
		(0.65459673E1*np.log(TEMP))))
	#print(Psat_water)
	# convert saturation vapour pressures from Pa to molecules/cc (air) using ideal
	# gas law, R has units cc.Pa/K.mol
	H2Oc = (RH*Psat_water)*(NA/(8.3144598e6*TEMP))
	
	# convert Psat_water to log10(atm) from Pa to be consistent with vapour pressures 
	# in volat_calc later on
	Psat_water = np.log10(Psat_water/101325.0)
	H2O_mw = 18.0 # state molecular weight of water (g/mol)
	
	return(H2Oc, Psat_water, H2O_mw)



'''calls modules to setup simulation'''
# the modules necessary to setup a simulation are called here

import eqn_pars
import user_input as ui
import ode_updater
import init_conc
import prop_calc
import partit_var_prep
import pp_intro
import time
import save
import pickle
import os

# define function
def middle():
	
	# get required inputs
	[sav_nam, sch_name, chem_sch_mrk, xml_name, update_stp, tot_time, 
		comp0, y0, temp, tempt, RH, Pnow, wall_on,
		Cw, kw, siz_str, num_sb, pmode, pconc, pconct,
		lowsize, uppsize, space_mode, std, mean_rad,
		save_step, const_comp, Compt, injectt, Ct, seed_name, seed_mw, 
		core_diss, core_dens, seedVr, light_stat, light_time, daytime, lat, 
		lon, af_path, dayOfYear, photo_path, tf, light_ad, con_infl_nam, 
		const_infl_t, con_infl_C, dydt_trak, dens_comp, dens, vol_comp, 
		volP, act_comp, act_user, accom_comp, accom_coeff_user, uman_up, 
		int_tol, new_partr, nucv1, nucv2, nucv3, nuc_comp, nuc_ad, coag_on, 
		inflectDp, pwl_xpre, pwl_xpro, inflectk, ChamR, Rader, p_char, 
		e_field, dil_fac, partit_cutoff, ser_H2O] = ui.share(0)
	
	# parse the chemical scheme equation file to convert equations
	# into usable code
	[rindx_g, pindx_g, rstoi_g, pstoi_g, nreac_g, nprod_g, 
	jac_stoi_g, njac_g, jac_den_indx_g, 
	jac_indx_g, y_arr_g, y_rind_g, uni_y_rind_g, y_pind_g, uni_y_pind_g, 
	reac_col_g, prod_col_g, rstoi_flat_g, 
	pstoi_flat_g, rr_arr_g, rr_arr_p_g, rowvals, colptrs, jac_wall_indx, 
	jac_part_indx, comp_num, RO2_indx, comp_list, Pybel_objects, eqn_num, 
	comp_namelist, Jlen, 
	rindx_aq, rstoi_aq, pindx_aq, pstoi_aq, reac_coef_aq, 
	nreac_aq, nprod_aq, jac_stoi_aq, 
	jac_den_indx_aq, njac_aq, jac_indx_aq, 				
	y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, 
	uni_y_pind_aq, reac_col_aq, prod_col_aq, rstoi_flat_aq, pstoi_flat_aq, 
	rr_arr_aq, rr_arr_p_aq, comp_xmlname, comp_smil] = eqn_pars.extr_mech(sch_name, 
	chem_sch_mrk, xml_name, photo_path, con_infl_nam, int_tol, wall_on, 
	(num_sb+wall_on), const_comp)
	
	# set initial concentrations (molecules/cc)
	[y, H2Oi, y_mw, num_comp, Cfactor, indx_plot, corei, dydt_vst, comp_namelist, 
	inj_indx, core_diss, Psat_water, 
	nuci, nrec_steps, seedi] = init_conc.init_conc(comp_num, comp0, y0, temp[0], RH, 
	Pnow, Pybel_objects, 0, pconc, dydt_trak, tot_time, save_step, rindx_g, 
	pindx_g, eqn_num[0], nreac_g, nprod_g, 
	comp_namelist, Compt, seed_name,
	seed_mw, core_diss, nuc_comp, comp_xmlname, comp_smil)
	
	
	# dump new pickle file ready for plotting script to use
	list_vars = [sav_nam, sch_name, indx_plot, comp0]
	input_by_sim = str(os.getcwd() + '/PyCHAM/pickle.pkl')
	with open(input_by_sim, 'wb') as f: # the file to be used for pickling
		pickle.dump(list_vars, f) # pickle
		f.close() # close

	# get component properties
	[Psat, y_dens, Psat_Pa] = prop_calc.prop_calc(comp_list, Pybel_objects, temp[0], H2Oi, 
		num_comp, Psat_water, vol_comp, volP, 0, corei, pconc,
		uman_up, core_dens, comp_namelist, 0, nuci, nuc_comp, num_sb, dens_comp, dens,
		seed_name)

	# prepare for the calcuation of partitioning variables
	[mfp, accom_coeff, therm_sp, surfT, Cw, act_coeff, 
		R_gas, NA, diff_vol, Dstar_org] = partit_var_prep.prep(y_mw, 
		temp[0], num_comp, 0, Cw, act_comp, act_user, accom_comp, 
		accom_coeff_user, comp_namelist, num_sb, num_sb, Pnow, 
		Pybel_objects, comp_smil)
	count = 0
	
	# prepare particle phase and wall
	[y, N_perbin, x, Varr, Vbou, rad0, Vol0, rbou, MV, num_sb, nuc_comp, 
	rbou00, ub_rad_amp, np_sum] = pp_intro.pp_intro(y, num_comp, Pybel_objects, temp[0],
	 H2Oi, mfp, accom_coeff, y_mw, surfT, RH, siz_str, num_sb, lowsize, 
		uppsize, pmode, pconc, pconct, nuc_comp, 0, std, mean_rad, 
		therm_sp, y_dens, Psat, core_diss, kw, space_mode, seedVr,
		comp_namelist, act_coeff, wall_on, partit_cutoff, Pnow, seedi)
	
	print('Calling integration routine, starting timer')
	st_time = time.time()
	
	# solve problem
	[trec, yrec, dydt_vst, Cfactor_vst, Nres_dry, Nres_wet, x2, rbou_rec] = ode_updater.ode_updater(update_stp, 
		tot_time, save_step, y, rindx_g, 
		pindx_g, rstoi_g, pstoi_g, nreac_g, nprod_g, jac_stoi_g, njac_g, 
		jac_den_indx_g, jac_indx_g, RO2_indx, H2Oi, temp, tempt, 
		Pnow, light_stat, light_time, daytime, lat, lon, af_path, 
		dayOfYear, photo_path, Jlen, con_infl_C, nrec_steps, 
		dydt_vst, siz_str, num_sb, num_comp, seedi, seed_name, seedVr, 
		core_diss, Psat, mfp, therm_sp,  
		accom_coeff, y_mw, surfT, R_gas, NA, y_dens, 
		x, Varr, act_coeff, Cw, kw, Cfactor, tf, light_ad, y_arr_g,
		y_rind_g, 
		uni_y_rind_g, y_pind_g, uni_y_pind_g, reac_col_g, prod_col_g, 
		rstoi_flat_g, pstoi_flat_g, rr_arr_g, rr_arr_p_g, rowvals, 
		colptrs, wall_on, jac_wall_indx, jac_part_indx, Vbou, 
		N_perbin, Vol0, rad0, np_sum, new_partr, nucv1, nucv2, 
		nucv3, nuc_comp, nuc_ad, RH, coag_on, inflectDp, pwl_xpre, 
		pwl_xpro, inflectk, ChamR, Rader, p_char, e_field, 
		injectt, inj_indx, Ct, pmode, pconc, pconct, mean_rad, lowsize, 
		uppsize, std, rbou, const_infl_t, MV,
		rindx_aq, 
		pindx_aq, rstoi_aq, pstoi_aq, nreac_aq, nprod_aq, jac_stoi_aq, njac_aq, 
		jac_den_indx_aq, jac_indx_aq, y_arr_aq,
		y_rind_aq, 
		uni_y_rind_aq, y_pind_aq, uni_y_pind_aq, reac_col_aq, prod_col_aq, 
		rstoi_flat_aq, pstoi_flat_aq, rr_arr_aq, rr_arr_p_aq, eqn_num,
		partit_cutoff, diff_vol, Dstar_org, corei, ser_H2O)
	
	time_taken = time.time()-st_time
	print('Simulation complete, wall clock time elapsed since first call to solver: ', time_taken, ' s')		

	print('Saving results')
	# save results
	save.saving(sch_name, yrec, Nres_dry, Nres_wet, trec, sav_nam, 
		dydt_vst, num_comp, Cfactor_vst, 0, 
		num_sb, comp_namelist, dydt_trak, y_mw, MV, time_taken, 
		seed_name, x2, rbou_rec, wall_on, space_mode, rbou00, ub_rad_amp)
	print('Saved results')
	
	
	return()



'''solution of ODEs, generated by eqn_pars.py'''
# module to solve system of ordinary differential equations (ODEs) using solve_ivp of Scipy 
# File Created at 2020-12-15 18:13:17.799277

import numpy as np
import scipy.sparse as SP
from scipy.integrate import solve_ivp

# define function
def ode_solv(y, integ_step, rindx, pindx, rstoi, pstoi, 
	nreac, nprod, rrc, jac_stoi, njac, jac_den_indx, 
	jac_indx, Cinfl_now, y_arr, y_rind, uni_y_rind, 
	y_pind, uni_y_pind, reac_col, prod_col, 
	rstoi_flat, pstoi_flat, rr_arr, rr_arr_p,
	rowvals, colptrs, num_comp, num_sb,
	wall_on, Psat, Cw, act_coeff, kw, jac_wall_indx,
	seedi, core_diss, kelv_fac, kimt, num_asb,
	jac_part_indx,
	rindx_aq, pindx_aq, rstoi_aq, pstoi_aq,
	nreac_aq, nprod_aq, jac_stoi_aq, njac_aq, jac_den_indx_aq, jac_indx_aq,
	y_arr_aq, y_rind_aq, uni_y_rind_aq, y_pind_aq, uni_y_pind_aq,
	reac_col_aq, prod_col_aq, rstoi_flat_aq,
	pstoi_flat_aq, rr_arr_aq, rr_arr_p_aq, eqn_num, jac_mod_len,
	jac_part_hmf_indx, rw_indx, N_perbin, jac_part_H2O_indx, H2Oi):

	# inputs: -------------------------------------
	# y - initial concentrations (moleucles/cm3)
	# integ_step - the maximum integration time step (s)
	# rindx - index of reactants per equation
	# pindx - index of products per equation
	# rstoi - stoichiometry of reactants
	# pstoi - stoichiometry of products
	# nreac - number of reactants per equation
	# nprod - number of products per equation
	# rrc - reaction rate coefficient
	# jac_stoi - stoichiometries relevant to Jacobian
	# njac - number of Jacobian elements affected per equation
	# jac_den_indx - index of component denominators for Jacobian
	# jac_indx - index of Jacobian to place elements per equation (rows)
	# Cinfl_now - influx of components with constant influx 
	#		(molecules/cc/s)
	# y_arr - index for matrix used to arrange concentrations, 
	#	enabling calculation of reaction rate coefficients 
	# y_rind - index of y relating to reactants for reaction rate 
	# 	coefficient equation
	# uni_y_rind - unique index of reactants 
	# y_pind - index of y relating to products
	# uni_y_pind - unique index of products 
	# reac_col - column indices for sparse matrix of reaction losses
	# prod_col - column indices for sparse matrix of production gains
	# rstoi_flat - 1D array of reactant stoichiometries per equation
	# pstoi_flat - 1D array of product stoichiometries per equation
	# rr_arr - index for reaction rates to allow reactant loss
	# 	calculation
	# rr_arr_p - index for reaction rates to allow reactant loss
	# 	calculation
	# rowvals - row indices of Jacobian elements
	# colptrs - indices of  rowvals corresponding to each column of the
	# 	Jacobian
	# num_comp - number of components
	# num_sb - number of size bins
	# wall_on - flag saying whether to include wall partitioning
	# Psat - pure component saturation vapour pressures (molecules/cc)
	# Cw - effective absorbing mass concentration of wall (molecules/cc) 
	# act_coeff - activity coefficient of components
	# kw - mass transfer coefficient to wall (/s)
	# jac_wall_indx - index of inputs to Jacobian by wall partitioning
	# seedi - index of seed material
	# core_diss - dissociation constant of seed material
	# kelv_fac - kelvin factor for particles
	# kimt - mass transfer coefficient for gas-particle partitioning (s)
	# num_asb - number of actual size bins (excluding wall)
	# jac_part_indx - index for sparse Jacobian for particle influence 
	# eqn_num - number of gas- and aqueous-phase reactions 
	# jac_mod_len - modification length due to high fraction of component(s)
	# 	in particle phase
	# jac_part_hmf_indx - index of Jacobian affected by water
	#	 in the particle phase
	# rw_indx - indices of rows affected by water in particle phase
	# N_perbin - number concentration of particles per size bin (#/cc)
	# jac_part_H2O_indx - sparse Jacobian indices for the effect of
	#	particle-phase water on all other components
	# H2Oi - index for water
	# ---------------------------------------------

	def dydt(t, y): # define the ODE(s)
		
		# inputs: ----------------
		# y - concentrations (molecules/cc), note when using scipy integrator solve_ivp, this should have shape (number of elements, 1)
		# t - time interval to integrate over (s)
		# ---------------------------------------------
		
		# ensure y is correct shape
		if (y.shape[1] > 1):
			y = y[:, 0].reshape(-1, 1)
		# empty array to hold rate of change per component
		dd = np.zeros((y.shape[0], 1))
		
		# gas-phase reactions -------------------------
		# empty array to hold relevant concentrations for
		# reaction rate coefficient calculation
		rrc_y = np.ones((rindx.shape[0]*rindx.shape[1]))
		rrc_y[y_arr] = y[y_rind, 0]
		rrc_y = rrc_y.reshape(rindx.shape[0], rindx.shape[1], order = 'C')
		# reaction rate (molecules/cc/s) 
		rr = rrc[0:rindx.shape[0]]*((rrc_y**rstoi).prod(axis=1))
		# loss of reactants
		data = rr[rr_arr]*rstoi_flat # prepare loss values
		# convert to sparse matrix
		loss = SP.csc_matrix((data, y_rind, reac_col))
		# register loss of reactants
		dd[uni_y_rind, 0] -= np.array((loss.sum(axis = 1))[uni_y_rind])[:, 0]
		# gain of products
		data = rr[rr_arr_p]*pstoi_flat # prepare loss values
		# convert to sparse matrix
		loss = SP.csc_matrix((data, y_pind, prod_col))
		# register gain of products
		dd[uni_y_pind, 0] += np.array((loss.sum(axis = 1))[uni_y_pind])[:, 0]
		
		# gas-wall partitioning ----------------
		# concentration on wall (molecules/cc (air))
		Csit = y[num_comp*(num_asb+1):num_comp*(num_asb+2), 0]
		# saturation vapour pressure on wall (molecules/cc (air))
		# note, just using the top rows of Psat and act_coeff
		# as do not need the repetitions over size bins
		if (Cw > 0.):
			Csit = Psat[0, :]*(Csit/Cw)*act_coeff[0, :]
			# rate of transfer (molecules/cc/s)
			dd_all = kw*(y[0:num_comp, 0]-Csit)
			dd[0:num_comp, 0] -= dd_all # gas-phase change
			dd[num_comp*num_sb:num_comp*(num_sb+1), 0] += dd_all # wall change
		
		
		dd = (dd[:, 0]).reshape(num_sb+1, num_comp)
		# force all components in size bins with no particle to zero
		if (num_asb > 0):
			dd[1:num_asb+1, :][N_perbin[:, 0] == 0, :] = 0
		# return to array, note that consistent with the solve_ivp manual, this ensures dd is
		# a vector rather than matrix, since y0 is a vector
		dd = dd.flatten()
		return (dd)

	def jac(t, y): # define the Jacobian
		
		# inputs: ----------------
		# y - concentrations (molecules/cc), note when using scipy integrator solve_ivp, this should have shape (number of elements, 1)
		# t - time interval to integrate over (s)
		# ---------------------------------------------
		
		# ensure y is correct shape
		if (y.ndim == 2):
			if (y.shape[1] > 1):
				y = y[:, 0].reshape(-1, 1)
		if (y.ndim <= 1):
			y = y.reshape(-1, 1)
		
		# elements of sparse Jacobian matrix
		data = np.zeros((102))
		
		for i in range(rindx.shape[0]): # gas-phase reaction loop
			# reaction rate (molecules/cc/s)
			rr = rrc[i]*(y[rindx[i, 0:nreac[i]], 0].prod())
			# prepare Jacobian inputs
			jac_coeff = np.zeros((njac[i, 0]))
			# only fill Jacobian if reaction rate sufficient
			if (rr != 0.):
				jac_coeff = (rr*(jac_stoi[i, 0:njac[i, 0]])/
				(y[jac_den_indx[i, 0:njac[i, 0]], 0]))
			data[jac_indx[i, 0:njac[i, 0]]] += jac_coeff
		
		
		if (Cw > 0.):
			wall_eff = np.zeros((76))
			wall_eff[0:38:2] = -kw # effect of gas on gas 
			wall_eff[1:38:2] = +kw # effect of gas on wall 
			# effect of wall on gas
			wall_eff[38:76:2] = +kw*(Psat[0,:]*act_coeff[0, :]/Cw) 
			# effect of wall on wall
			wall_eff[38+1:76:2] = -kw*(Psat[0,:]*act_coeff[0, :]/Cw) 
			data[jac_wall_indx] += wall_eff
		
		# create Jacobian
		j = SP.csc_matrix((data, rowvals, colptrs))
		
		return(j)
	
	# set ODE solver tolerances
	atol = 0.001
	rtol = 0.0001
	
	# call on the ODE solver, note y contains the initial condition(s) (molecules/cc (air)) and must be 1D even though y in dydt and jac has shape (number of elements, 1)
	sol = solve_ivp(dydt, [0, integ_step], y, atol = atol, rtol = rtol, method = 'BDF', t_eval = [integ_step], vectorized = True, jac = jac)
	
	# force all components in size bins with no particle to zero
	y = np.squeeze(sol.y)
	y = y.reshape(num_sb+1, num_comp)
	if (num_asb > 0):
		y[1:num_asb+1, :][N_perbin[:, 0] == 0, :] = 0
	# return to array
	y = y.flatten()
	
	# return concentration(s) and time(s) following integration
	return(y, sol.t)
